# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import logging
import os
import warnings
from collections import defaultdict
from collections.abc import Callable
from collections.abc import Iterator
from functools import cached_property
from types import MappingProxyType as frozendict
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal

import deprecation
import earthkit.data as ekd
import numpy as np
from anemoi.transform.variables import Variable
from anemoi.utils.config import DotDict
from anemoi.utils.dates import frequency_to_timedelta as to_timedelta
from anemoi.utils.provenance import gather_provenance_info

from anemoi.inference._version import __version__
from anemoi.inference.types import DataRequest
from anemoi.inference.types import FloatArray
from anemoi.inference.types import IntArray

from .legacy import LegacyMixin
from .patch import PatchMixin

if TYPE_CHECKING:
    from earthkit.data import FieldList

USE_LEGACY = True

LOG = logging.getLogger(__name__)


VARIABLE_CATEGORIES = {
    "computed",
    "forcing",
    "diagnostic",
    "prognostic",
    "constant",
    "accumulation",
}


def _remove_full_paths(x: Any) -> Any:
    """Remove full paths from the given data structure.

    Parameters
    ----------
    x : Any
        The data structure to process.

    Returns
    -------
    Any
        The processed data structure with full paths removed.
    """
    if isinstance(x, dict):
        return {k: _remove_full_paths(v) for k, v in x.items()}
    if isinstance(x, list):
        return [_remove_full_paths(v) for v in x]
    if isinstance(x, str):
        return os.path.splitext(os.path.basename(x))[0]
    return x


class Metadata(PatchMixin, LegacyMixin):
    """An object that holds metadata of a checkpoint."""

    def __init__(self, metadata: dict[str, Any], supporting_arrays: dict[str, FloatArray] = {}):
        """Initialize the Metadata object.

        Parameters
        ----------
        metadata : dict
            The metadata dictionary.
        supporting_arrays : dict, optional
            The supporting arrays, by default {}.
        """
        self._metadata = DotDict(metadata)
        assert isinstance(supporting_arrays, dict)
        self._supporting_arrays = supporting_arrays
        self._variables_categories = None

    @property
    def _indices(self) -> DotDict:
        """Return the data indices."""
        return self._metadata.data_indices

    @property
    def _config_data(self) -> DotDict:
        """Return the data configuration."""
        return self._config.data

    @property
    def _config_training(self) -> DotDict:
        """Return the training configuration."""
        return self._config.training

    @property
    def _config_model(self) -> DotDict:
        """Return the model configuration."""
        return self._config.model

    @property
    def _config(self) -> DotDict:
        """Return the configuration."""
        return self._metadata.config

    @property
    def target_explicit_times(self) -> Any:
        """Return the target explicit times from the training configuration."""
        return self._config_training.explicit_times.target

    @property
    def input_explicit_times(self) -> Any:
        """Return the input explicit times from the training configuration."""
        return self._config_training.explicit_times.input

    ###########################################################################
    # Debugging
    ###########################################################################

    def _print_indices(
        self, title: str, indices: dict[str, list[int]], naming: dict, skip: list[str] = [], print=LOG.info
    ) -> None:
        """Print indices for debugging purposes.

        Parameters
        ----------
        title : str
            The title for the indices.
        indices : Dict[str, List[int]]
            The indices to print.
        naming : dict
            The naming convention for the indices.
        skip : set, optional
            The set of indices to skip, by default set().
        print : callable, optional
            The print function to use, by default logging.info.
        """
        print("")
        print(title)

        for k, v in sorted(indices.items()):
            if k in skip:
                continue
            for name, idx in sorted(v.items()):
                entry = f"{k}.{name}"
                if entry in skip:
                    continue

                print(f"   {k}.{name}:")
                for n in idx:
                    print(f"     {n:3d} - {naming[k].get(n, '?')}")
                if not idx:
                    print("     <empty>")

    def print_indices(self, print=LOG.info) -> None:
        """Print data and model indices for debugging purposes."""
        v = {i: v for i, v in enumerate(self.variables)}
        r = {v: k for k, v in self.variable_to_input_tensor_index.items()}
        s = self.output_tensor_index_to_variable

        self._print_indices("Data indices", self._indices.data, dict(input=v, output=v), skip=["output"], print=print)
        self._print_indices(
            "Model indices", self._indices.model, dict(input=r, output=s, skip=["output.full"]), print=print
        )

    ###########################################################################
    # Inference
    ###########################################################################

    @cached_property
    def timestep(self) -> datetime.timedelta:
        """Model time stepping timestep."""
        # frequency = to_timedelta(self._config_data.frequency)
        timestep = to_timedelta(self._config_data.timestep)
        return timestep

    @cached_property
    def precision(self) -> str | int:
        """Return the precision of the model (bits per float)."""
        return self._config_training.precision

    def _make_indices_mapping(self, indices_from: list, indices_to: list) -> frozendict:
        """Create a mapping between two sets of indices.

        Parameters
        ----------
        indices_from : list
            The source indices.
        indices_to : list
            The target indices.

        Returns
        -------
        frozendict
            The mapping between the source and target indices.
        """
        assert len(indices_from) == len(indices_to), (indices_from, indices_to)
        return frozendict({i: j for i, j in zip(indices_from, indices_to)})

    @property
    def variable_to_input_tensor_index(self) -> frozendict:
        """Return the mapping between variable name and input tensor index."""
        mapping = self._make_indices_mapping(
            self._indices.data.input.full,
            self._indices.model.input.full,
        )

        return frozendict({v: mapping[i] for i, v in enumerate(self.variables) if i in mapping})

    @cached_property
    def output_tensor_index_to_variable(self) -> frozendict:
        """Return the mapping between output tensor index and variable name."""
        mapping = self._make_indices_mapping(
            self._indices.model.output.full,
            self._indices.data.output.full,
        )
        return frozendict({k: self.variables[v] for k, v in mapping.items()})

    @cached_property
    def number_of_grid_points(self) -> int:
        """Return the number of grid points per fields."""
        if "grid_indices" in self._supporting_arrays:
            return len(self.load_supporting_array("grid_indices"))
        try:
            return self._metadata.dataset.shape[-1]
        except AttributeError:
            if not USE_LEGACY:
                raise
            return self._legacy_number_of_grid_points()

    @cached_property
    def number_of_input_features(self) -> int:
        """Return the number of input features."""
        return len(self._indices.model.input.full)

    @cached_property
    def model_computed_variables(self) -> tuple:
        """The initial conditions variables that need to be computed and not retrieved."""
        typed_variables = self.typed_variables
        return tuple(name for name, v in typed_variables.items() if v.is_computed_forcing)

    @cached_property
    def multi_step_input(self) -> int:
        """Number of past steps needed for the initial conditions tensor."""
        return self._config_training.multistep_input

    @cached_property
    def prognostic_output_mask(self) -> IntArray:
        """Return the prognostic output mask."""
        return np.array(self._indices.model.output.prognostic)

    @cached_property
    def prognostic_input_mask(self) -> IntArray:
        """Return the prognostic input mask."""
        return np.array(self._indices.model.input.prognostic)

    @property
    @deprecation.deprecated(
        deprecated_in="0.6.4",
        removed_in="0.7.0",
        current_version=__version__,
        details="Use `select_variables_and_mask` instead.",
    )
    def computed_time_dependent_forcings(self) -> tuple[np.ndarray, list]:
        """Return the indices and names of the computed forcings that are not constant in time."""
        # Mapping between model and data indices
        mapping = self._make_indices_mapping(
            self._indices.model.input.full,
            self._indices.data.input.full,
        )

        # Mapping between model indices and variable names
        forcings_variables = {self.variables[mapping[i]]: i for i in self._indices.model.input.forcing}
        typed_variables = self.typed_variables

        # Filter out the computed forcings that are not constant in time
        indices = []
        variables = []
        for name, idx in sorted(forcings_variables.items(), key=lambda x: x[1]):
            v = typed_variables[name]
            if v.is_computed_forcing and not v.is_constant_in_time:
                indices.append(idx)
                variables.append(name)

        return np.array(indices), variables

    @property
    @deprecation.deprecated(
        deprecated_in="0.6.4",
        removed_in="0.7.0",
        current_version=__version__,
        details="Use `select_variables_and_mask` instead.",
    )
    def computed_constant_forcings(self) -> tuple[FloatArray, list[str]]:
        """Return the indices and names of the computed forcings that are  constant in time."""
        # Mapping between model and data indices
        mapping = self._make_indices_mapping(
            self._indices.model.input.full,
            self._indices.data.input.full,
        )

        # Mapping between model indices and variable names
        forcings_variables = {self.variables[mapping[i]]: i for i in self._indices.model.input.forcing}
        typed_variables = self.typed_variables

        # Filter out the computed forcings that are not constant in time
        indices = []
        variables = []
        for name, idx in sorted(forcings_variables.items(), key=lambda x: x[1]):
            v = typed_variables[name]
            if v.is_computed_forcing and v.is_constant_in_time:
                indices.append(idx)
                variables.append(name)

        return np.array(indices), variables

    def has_supporting_array(self, name: str) -> bool:
        """Check if the metadata has a supporting array with the given name.

        Parameters
        ----------
        name : str
            The name of the supporting array.

        Returns
        -------
        bool
            True if the supporting array exists, False otherwise.
        """
        return name in self._supporting_arrays

    ###########################################################################
    # Variables
    ###########################################################################

    @property
    def variables(self) -> tuple:
        """Return the variables as found in the training dataset."""
        return tuple(self._metadata.dataset.variables)

    @cached_property
    def variables_metadata(self) -> dict[str, Any]:
        """Return the variables and their metadata as found in the training dataset."""
        try:
            result = self._metadata.dataset.variables_metadata
            if USE_LEGACY:
                self._legacy_check_variables_metadata(result)
        except AttributeError:
            if not USE_LEGACY:
                raise
            result = self._legacy_variables_metadata()

        if "constant_fields" in self._metadata.dataset:
            for name in self._metadata.dataset.constant_fields:
                if name not in result:
                    continue
                result[name]["constant_in_time"] = True

        return result

    @cached_property
    @deprecation.deprecated(
        deprecated_in="0.6.4",
        removed_in="0.7.0",
        current_version=__version__,
        details="Use `select_variables` instead.",
    )
    def diagnostic_variables(self) -> list:
        """Variables that are marked as diagnostic."""
        return [self.index_to_variable[i] for i in self._indices.data.input.diagnostic]

    @cached_property
    @deprecation.deprecated(
        deprecated_in="0.6.4",
        removed_in="0.7.0",
        current_version=__version__,
        details="Use `select_variables` instead.",
    )
    def prognostic_variables(self) -> list:
        """Variables that are marked as prognostic."""
        return [self.index_to_variable[i] for i in self._indices.data.input.prognostic]

    @cached_property
    def index_to_variable(self) -> frozendict:
        """Return a mapping from index to variable name."""
        return frozendict({i: v for i, v in enumerate(self.variables)})

    @cached_property
    def typed_variables(self) -> dict[str, Variable]:
        """Returns a strongly typed variables."""
        result = {name: Variable.from_dict(name, self.variables_metadata[name]) for name in self.variables}

        if "cos_latitude" in result:
            assert result["cos_latitude"].is_computed_forcing
            assert result["cos_latitude"].is_constant_in_time

        if "cos_julian_day" in result:
            assert result["cos_julian_day"].is_computed_forcing
            assert not result["cos_julian_day"].is_constant_in_time

        return result

    @cached_property
    def accumulations(self) -> list:
        """Return the indices of the variables that are accumulations."""
        return [v.name for v in self.typed_variables.values() if v.is_accumulation]

    def name_fields(self, fields: ekd.FieldList, namer: Callable[..., str] | None = None) -> "FieldList":
        """Name fields using the provided namer.

        Parameters
        ----------
        fields : FieldList
            The fields to name.
        namer : callable, optional
            The namer function, by default None.

        Returns
        -------
        FieldList
            The named fields.
        """
        from earthkit.data.indexing.fieldlist import FieldArray

        if namer is None:
            namer = self.default_namer()

        def _name(field: ekd.Field, _: str, original_metadata: dict[str, Any]) -> str:
            return namer(field, original_metadata)

        return FieldArray([f.clone(name=_name) for f in fields])

    def sort_by_name(
        self,
        fields: ekd.FieldList,
        *args: Any,
        namer: Callable[..., Any] | None = None,
        **kwargs: Any,
    ) -> ekd.FieldList:
        """Sort fields by name.

        Parameters
        ----------
        fields : ekd.FieldList
            The fields to sort.
        args : Any
            Additional arguments.
        namer : callable, optional
            The namer function, by default None.
        kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        ekd.FieldList
            The sorted fields.
        """
        fields = self.name_fields(fields, namer=namer)
        return fields.order_by("name", *args, **kwargs)

    ###########################################################################
    # Default namer
    ###########################################################################

    def default_namer(self, *args: Any, **kwargs: Any) -> Callable[..., str]:
        """Return a callable that can be used to name earthkit-data fields.

        Parameters
        ----------
        args : Any
            Additional arguments.
        kwargs : Any
            Additional keyword arguments.

        Returns
        -------
        Callable
            The namer function.
        """
        assert len(args) == 0, args
        assert len(kwargs) == 0, kwargs

        def namer(field: ekd.Field, metadata: dict[str, Any]) -> str:
            # TODO: Return the `namer` used when building the dataset
            warnings.warn("🚧  TEMPORARY CODE 🚧: Use the remapping in the metadata")
            param, levelist, levtype = (
                metadata.get("param"),
                metadata.get("levelist"),
                metadata.get("levtype"),
            )

            # Bug in eccodes that returns levelist for single level fields in GRIB2
            if levtype in ("sfc", "o2d"):
                levelist = None

            if levelist is None:
                return param

            return f"{param}_{levelist}"

        return namer

    ###########################################################################
    # Data retrieval
    ###########################################################################

    @property
    def _data_request(self) -> DataRequest:
        """Return the data request as encoded in the dataset."""
        try:
            return self._metadata.dataset.data_request
        except AttributeError:
            return self._legacy_data_request()

    @property
    def grid(self) -> str | None:
        """Return the grid information."""
        return self._data_request.get("grid")

    @property
    def area(self) -> str | None:
        """Return the area information."""
        return self._data_request.get("area")

    def select_variables(
        self,
        *,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        has_mars_requests: bool = True,
    ) -> list[str]:
        """Get variables from input.

        Parameters
        ----------
        include: List[str]
            Categories to include.
        exclude: List[str]
            Categories to exclude.
        has_mars_requests: bool
            If True, only include variables that have MARS requests.

        Returns
        -------
        List[str]
            The list of variables.
        """

        variable_categories = self.variable_categories()
        result = []

        def parse_category(categories: str) -> set:
            if categories is None:
                return None

            parsed = set()
            for category in categories:
                bits = category.split("+")
                for bit in bits:
                    if bit not in VARIABLE_CATEGORIES:
                        raise ValueError(f"Unknown category {bit} in {categories}. Known: {VARIABLE_CATEGORIES}")
                parsed.add(frozenset(bits))
            return parsed

        def match(categories, variable_categories, variable):
            for c in categories:
                if c.issubset(variable_categories):
                    return True
            return False

        include = parse_category(include)
        exclude = parse_category(exclude)

        if include is not None and exclude is not None:
            if not include.isdisjoint(exclude):
                raise ValueError(f"Include and exclude sets must not overlap {include} & {exclude}")

        for variable, metadata in self.variables_metadata.items():

            categories = set(variable_categories[variable])

            if not categories < VARIABLE_CATEGORIES:
                warnings.warn(
                    f"Variable {variable} has unknown categories: {categories - VARIABLE_CATEGORIES}. "
                    f"Please update the code."
                )

            if has_mars_requests and "mars" not in metadata:
                continue

            if exclude is not None and match(exclude, categories, variable):
                continue

            if include is None or match(include, categories, variable):
                result.append(variable)

        return result

    def variables_mask(self, *, variables: list[str]) -> IntArray:

        variable_to_input_tensor_index = self.variable_to_input_tensor_index
        indices = [variable_to_input_tensor_index[v] for v in variables]

        return np.array(indices)

    def select_variables_and_masks(
        self, *, include: list[str] | None = None, exclude: list[str] | None
    ) -> tuple[list[str], IntArray]:
        variables = self.select_variables(include=include, exclude=exclude, has_mars_requests=False)
        return variables, self.variables_mask(variables=variables)

    def mars_input_requests(self) -> Iterator[DataRequest]:
        """Generate MARS input requests.

        Returns
        -------
        Iterator[DataRequest]
            The MARS requests.
        """

        for variable in self.select_variables(include=["prognostic", "forcing"], exclude=["computed", "diagnostic"]):
            metadata = self.variables_metadata[variable]

            yield metadata["mars"].copy()

    def mars_by_levtype(self, levtype: str) -> tuple[set, set]:
        """Get MARS parameters and levels by levtype.

        Parameters
        ----------
        levtype : str
            The levtype to filter by.

        Returns
        -------
        tuple
            The parameters and levels.
        """

        params = set()
        levels = set()

        for variable in self.select_variables(
            include=["prognostic", "forcing"],
            exclude=["computed", "diagnostic"],
        ):

            metadata = self.variables_metadata[variable]

            mars = metadata["mars"]
            if mars.get("levtype") != levtype:
                continue

            if "param" in mars:
                params.add(mars["param"])

            if "levelist" in mars:
                levels.add(mars["levelist"])

        return params, levels

    def mars_requests(self, *, variables: list[str]) -> Iterator[DataRequest]:
        """Generate MARS requests for the given variables.

        Parameters
        ----------
        variables : list
            The list of variables.

        Returns
        -------
        Iterator[DataRequest]
            The MARS requests.

        Raises
        ------
        ValueError
            If no variables are requested or if a variable is not found in the metadata.
        """
        if len(variables) == 0:
            raise ValueError("No variables requested")

        for variable in variables:

            if variable not in self.variables_metadata:
                raise ValueError(f"Variable {variable} not found in the metadata")

            if "mars" not in self.variables_metadata[variable]:
                raise ValueError(f"Variable {variable} has no MARS metadata")

            mars = self.variables_metadata[variable]["mars"].copy()

            for k in ("date", "hdate", "time", "valid_datetime", "variable"):
                mars.pop(k, None)

            yield mars

    ###########################################################################
    # Error reporting
    ###########################################################################

    def report_error(self) -> None:
        """Report an error with provenance information."""
        provenance = self._metadata.provenance_training

        def _print(title: str, provenance: dict[str, Any]) -> None:
            LOG.info("")
            LOG.info("%s:", title)
            for package, git in sorted(provenance.get("git_versions", {}).items()):
                if package.startswith("anemoi."):
                    sha1 = git.get("git", {}).get("sha1", "unknown")
                    LOG.info(f"   {package:20}: {sha1}")

            for package, version in sorted(provenance.get("module_versions", {}).items()):
                if package.startswith("anemoi."):
                    LOG.info(f"   {package:20}: {version}")

        _print("Environment used during training", provenance)
        _print("This environment", gather_provenance_info())

        LOG.warning("If you are running from a git repository, the versions reported above may not be accurate.")
        LOG.warning("The versions are only updated after a `pip install -e .`")

    def validate_environment(
        self,
        *,
        all_packages: bool = False,
        on_difference: Literal["warn", "error", "ignore", "return"] = "warn",
        exempt_packages: list[str] | None = None,
    ) -> bool | str:
        """Validate environment of the checkpoint against the current environment.

        Parameters
        ----------
        all_packages : bool, optional
            Check all packages in the environment (True) or just anemoi's (False), by default False.
        on_difference : Literal['warn', 'error', 'ignore', 'return'], optional
            What to do on difference, by default "warn"
        exempt_packages : list[str], optional
            List of packages to exempt from the check, by default EXEMPT_PACKAGES

        Returns
        -------
        Union[bool, str]
            boolean if `on_difference` is not 'return', otherwise formatted text of the differences
            True if environment is valid, False otherwise

        Raises
        ------
        RuntimeError
            If found difference and `on_difference` is 'error'
        ValueError
            If `on_difference` is not 'warn' or 'error'
        """
        from anemoi.inference.provenance import validate_environment

        return validate_environment(
            self,
            all_packages=all_packages,
            on_difference=on_difference,
            exempt_packages=exempt_packages,
        )

    ###########################################################################

    def _get_datasets_full_paths(self) -> list:
        """Get the full paths of the datasets used in the training.

        Returns
        -------
        list
            The list of full paths.
        """
        # This is a temporary method to get the full paths of the datasets used in the training.
        # we need to review what goes in the dataset metadata.

        result = []

        def _find(x: Any) -> None:

            if isinstance(x, list):
                for y in x:
                    if isinstance(y, str):
                        result.append(y)
                    else:
                        _find(y)

            if isinstance(x, dict):
                if "dataset" in x and isinstance(x["dataset"], str):
                    result.append(x["dataset"])

                for k, v in x.items():
                    _find(v)

        _find(self._config.dataloader.training.dataset)
        return result

    def open_dataset(
        self, *, use_original_paths: bool | None = None, from_dataloader: str | None = None
    ) -> tuple[Any, Any]:
        """Open the dataset.

        Parameters
        ----------
        use_original_paths : bool
            Whether to use the original paths.
        from_dataloader : str, optional
            The dataloader to use, by default None.

        Returns
        -------
        tuple
            The opened dataset and its arguments.
        """
        from anemoi.datasets import open_dataset
        from anemoi.utils.config import temporary_config

        if use_original_paths is not None:
            args, kwargs = self.open_dataset_args_kwargs(
                use_original_paths=use_original_paths, from_dataloader=from_dataloader
            )
            return open_dataset(*args, **kwargs)

        args, kwargs = self.open_dataset_args_kwargs(use_original_paths=True, from_dataloader=from_dataloader)

        #  Extract paths

        paths = []

        def _(x: Any) -> Any:
            if isinstance(x, dict):
                return {k: _(v) for k, v in x.items()}
            if isinstance(x, list):
                return [_(v) for v in x]

            if isinstance(x, str):
                if x.endswith(".zarr"):
                    paths.append(os.path.basename(x))
                    x = os.path.basename(x)
                    x = os.path.splitext(x)[0]

            return x

        args, kwargs = _((args, kwargs))

        with temporary_config(dict(datasets=dict(use_search_path_not_found=True))):
            return open_dataset(*args, **kwargs)

    def open_dataset_args_kwargs(
        self, *, use_original_paths: bool, from_dataloader: str | None = None
    ) -> tuple[Any, Any]:
        """Get the arguments and keyword arguments for opening the dataset.

        Parameters
        ----------
        use_original_paths : bool
            Whether to use the original paths.
        from_dataloader : str, optional
            The dataloader to use, by default None.

        Returns
        -------
        tuple
            The arguments and keyword arguments.
        """
        # Rebuild the full paths
        # Some older checkpoints may not have the full paths

        full_paths = self._get_datasets_full_paths()
        mapping = {}

        for path in full_paths:
            mapping[os.path.basename(path)] = path
            mapping[os.path.splitext(os.path.basename(path))[0]] = path

        def _fix(x: Any) -> Any:
            if isinstance(x, list):
                return [_fix(a) for a in x]

            if isinstance(x, dict):
                return {k: _fix(v) for k, v in x.items()}

            if isinstance(x, str):
                return mapping.get(x, x)

            return x

        if from_dataloader is not None:
            args, kwargs = [], self._metadata.config.dataloader[from_dataloader]
        else:
            args, kwargs = self._metadata.dataset.arguments.args, self._metadata.dataset.arguments.kwargs

        args, kwargs = _fix([args, kwargs])

        if use_original_paths:
            return args, kwargs

        return _remove_full_paths(args), _remove_full_paths(kwargs)

    ###########################################################################
    # Not sure this belongs here
    # We need factories
    ###########################################################################

    def variable_categories(self) -> dict:
        """Get the categories of variables.

        Returns
        -------
        dict
            The categories of variables.
        """

        if self._variables_categories is not None:
            return self._variables_categories

        result = defaultdict(set)
        typed_variables = self.typed_variables

        variables_in_data_space = self.variables
        variables_in_model_space = self.output_tensor_index_to_variable

        for name in self._config.data.forcing:
            result[name].add("forcing")

        for idx in self._indices.data.input.diagnostic:
            name = variables_in_data_space[idx]
            result[name].add("diagnostic")

        # assert self._indices.model.input.prognostic == self._indices.model.output.prognostic
        for idx in self._indices.model.output.prognostic:
            name = variables_in_model_space[idx]
            result[name].add("prognostic")

        for name, v in typed_variables.items():
            if v.is_accumulation:
                result[name].add("accumulation")

            if v.is_constant_in_time:
                result[name].add("constant")

            if v.is_computed_forcing:
                result[name].add("computed")

        for name in self.variables:
            if name not in result:
                raise ValueError(f"Variable {name} has no category")

            result[name] = sorted(result[name])

        self._variables_categories = frozendict(result)
        return self._variables_categories

    ###########################################################################
    # Supporting arrays
    ###########################################################################

    def load_supporting_array(self, name: str) -> FloatArray:
        """Load a supporting array by name.

        Parameters
        ----------
        name : str
            The name of the supporting array.

        Returns
        -------
        FloatArray
            The supporting array.

        Raises
        ------
        ValueError
            If the supporting array is not found.
        """
        if name not in self._supporting_arrays:
            LOG.error("No supporting array named `%s` found.", name)
            LOG.error("Supporting arrays found:")
            for names in self._supporting_arrays.keys():
                LOG.error("  %s", names)
            raise ValueError(f"Supporting array `{name}` not found")
        return self._supporting_arrays[name]

    @property
    def supporting_arrays(self) -> dict[str, FloatArray]:
        """Return the supporting arrays."""
        return self._supporting_arrays

    @property
    def latitudes(self) -> FloatArray | None:
        """Return the latitudes."""
        return self._supporting_arrays.get("latitudes")

    @property
    def longitudes(self) -> FloatArray | None:
        """Return the longitudes."""
        return self._supporting_arrays.get("longitudes")

    @property
    def grid_points_mask(self) -> FloatArray | None:
        """Return the grid points mask."""
        # TODO
        return None

    def provenance_training(self) -> dict[str, Any]:
        """Get the environmental configuration when trained.

        Returns
        -------
        dict
            The environmental configuration.
        """
        return dict(self._metadata.get("provenance_training", {}))

    def sources(self, path: str) -> list:
        """Get the sources from the metadata.

        Parameters
        ----------
        path : str
            The path to the sources.

        Returns
        -------
        list
            The list of sources.

        Raises
        ------
        ValueError
            If not all paths were fixed.
        """
        if "sources" not in self._metadata.dataset:
            return []

        import zipfile

        from anemoi.utils.checkpoints import load_supporting_arrays

        ###########################################################################
        # With older metadata, the zarr path are not stored in the metadata
        # we need to fix that
        ###########################################################################

        full_paths = self._get_datasets_full_paths()
        LOG.info(full_paths)
        n = 0

        def _fix(x: Any) -> Any:
            nonlocal n

            if isinstance(x, list):
                [_fix(a) for a in x]

            if isinstance(x, dict):
                if x.get("action", "").startswith("zarr"):
                    LOG.info(n, x)
                    path = full_paths[n]
                    n += 1
                    x["path"] = path

                {k: _fix(v) for k, v in x.items()}

        _fix(self._metadata.dataset.sources)

        if n != len(full_paths):
            raise ValueError("Not all paths were fixed")

        ###########################################################################

        sources = []

        with zipfile.ZipFile(path, "r") as zipf:
            for i, source in enumerate(self._metadata.dataset.get("sources", [])):
                entries = {
                    name: self._metadata.supporting_arrays_paths[name] for name in source.get("supporting_arrays", [])
                }
                arrays = load_supporting_arrays(zipf, entries)

                name = source.get("name")
                if name is None:
                    name = f"source{i}"

                sources.append(
                    SourceMetadata(
                        self,
                        name,
                        dict(dataset=source),
                        supporting_arrays=arrays,
                    )
                )

        return sources

    def print_variable_categories(self, print=LOG.info) -> None:
        """Print the variable categories for debugging purposes."""
        length = max(len(name) for name in self.variables)
        for name, categories in sorted(self.variable_categories().items()):
            print(f"   {name:{length}} => {', '.join(categories)}")

    ###########################################################################

    def patch(self, patch: dict) -> None:
        """Patch the metadata with the given patch.

        Parameters
        ----------
        patch : dict
            The patch to apply.
        """

        def merge(main: dict[str, Any], patch: dict[str, Any]) -> None:

            for k, v in patch.items():
                if isinstance(v, dict) and isinstance(main.get(k, {}), dict):
                    if k not in main:
                        main[k] = {}
                    merge(main[k], v)
                else:
                    main[k] = v

        merge(self._metadata, patch)


class SourceMetadata(Metadata):
    """An object that holds metadata of a source. It is only the `dataset` and `supporting_arrays` parts of the metadata.
    The rest is forwarded to the parent metadata object.
    """

    def __init__(self, parent: Metadata, name: str, metadata: dict, supporting_arrays: dict = {}):
        """Initialize the SourceMetadata object.

        Parameters
        ----------
        parent : Metadata
            The parent metadata object.
        name : str
            The name of the source.
        metadata : dict
            The metadata dictionary.
        supporting_arrays : dict, optional
            The supporting arrays, by default {}.
        """
        super().__init__(metadata, supporting_arrays)
        self.parent = parent
        self.name = name

    @property
    def latitudes(self) -> FloatArray | None:
        """Return the latitudes."""
        return self._supporting_arrays.get(f"{self.name}/latitudes")

    @property
    def longitudes(self) -> FloatArray | None:
        """Return the longitudes."""
        return self._supporting_arrays.get(f"{self.name}/longitudes")

    @property
    def grid_points_mask(self) -> FloatArray | None:
        """Return the grid points mask."""
        for k, v in self._supporting_arrays.items():
            # TODO: This is a bit of a hack
            if k.startswith(f"{self.name}/") and "mask" in k:
                return v
        return None

    ###########################################################################
    # Forward to parent metadata
    ###########################################################################

    @property
    def _config_training(self) -> DotDict:
        """Return the training configuration."""
        return self.parent._config_training

    @property
    def _config_data(self) -> DotDict:
        """Return the data configuration."""
        return self.parent._config_data

    @property
    def _indices(self) -> DotDict:
        """Return the indices."""
        return self.parent._indices

    @property
    def _config(self) -> DotDict:
        """Return the configuration."""
        return self.parent._config

    ###########################################################################



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from pathlib import Path

import numpy as np

from anemoi.inference.context import Context
from anemoi.inference.types import State
from anemoi.inference.utils.templating import render_template

from ..decorators import ensure_dir
from ..decorators import main_argument
from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("raw")
@main_argument("path")
@ensure_dir("dir")
class RawOutput(Output):
    """Raw output class."""

    def __init__(
        self,
        context: Context,
        dir: Path,
        template: str = "{date}.npz",
        strftime: str = "%Y%m%d%H%M%S",
        variables: list[str] | None = None,
        **kwargs,
    ) -> None:
        """Initialise the RawOutput class.

        Parameters
        ----------
        context : dict
            The context.
        dir : Path
            The directory to save the raw output.
            If the parent directory does not exist, it will be created.
        template : str, optional
            The template for filenames, by default "{date}.npz".
            Variables available are `date`, `basetime` `step`.
        strftime : str, optional
            The date format string, by default "%Y%m%d%H%M%S".
        """
        super().__init__(context, variables=variables, **kwargs)
        self.dir = dir
        self.template = template
        self.strftime = strftime

    def __repr__(self) -> str:
        """Return a string representation of the RawOutput object.

        Returns
        -------
        str
            String representation of the RawOutput object.
        """
        return f"RawOutput({self.dir})"

    def write_step(self, state: State) -> None:
        """Write the state to a compressed .npz file.

        Parameters
        ----------
        state : State
            The state to be written.
        """
        date = state["date"]
        basetime = date - state["step"]

        format_info = {
            "date": date.strftime(self.strftime),
            "step": state["step"],
            "basetime": basetime.strftime(self.strftime),
        }

        fn_state = f"{self.dir}/{render_template(self.template, format_info)}"
        restate = {f"field_{key}": val for key, val in state["fields"].items() if not self.skip_variable(key)}

        for key in ["date"]:
            restate[key] = np.array(state[key], dtype=str)

        for key in ["latitudes", "longitudes"]:
            restate[key] = np.array(state[key])

        np.savez_compressed(fn_state, **restate)



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import warnings

import numpy as np

from anemoi.inference.config import Configuration
from anemoi.inference.context import Context
from anemoi.inference.types import ProcessorConfig
from anemoi.inference.types import State

from ..output import ForwardOutput
from . import output_registry

LOG = logging.getLogger(__name__)


@output_registry.register("assign_mask")
class AssignMask(ForwardOutput):
    """Assign the current output to larger area using a mask.

    This operation can be seen as the opposite of "extract_mask".
    Instead of extracting a smaller area from a larger one,
    it assigns the current output to a larger area using a mask.
    This is useful when you want to restore the original state of the model
    after applying a mask to it. The portion of the state that is not
    covered by the mask will be set to a fill value (NaN by default).

    Parameters
    ----------
    context : dict
        The context dictionary.
    output : dict
        The output configuration dictionary.
    mask : str
        The mask supporting array name.
    fill_value : float, optional
        The fill value to use for the masked area, by default np.nan.
    post_processors : Optional[List[ProcessorConfig]], default None
        Post-processors to apply to the input
    output_frequency : int, optional
        The frequency of output, by default None.
    write_initial_state : bool, optional
        Whether to write the initial state, by default None.
    """

    def __init__(
        self,
        context: Context,
        *,
        output: Configuration,
        mask: str,
        fill_value: float = np.nan,
        post_processors: list[ProcessorConfig] | None = None,
        output_frequency: int | None = None,
        write_initial_state: bool | None = None,
    ) -> None:
        super().__init__(
            context, output, post_processors, output_frequency=output_frequency, write_initial_state=write_initial_state
        )
        warnings.warn(
            "The AssignMask output is deprecated and will be removed in a future release. "
            "Use the `assign_mask` post-processor instead.",
            DeprecationWarning,
        )

        if mask not in self.checkpoint.supporting_arrays:
            raise ValueError(f"Assignment mask '{mask}' not found in supporting arrays.")

        self.mask = self.checkpoint.load_supporting_array(mask)
        self.fill_value = fill_value

    def __repr__(self) -> str:
        """Return a string representation of the ExtractLamOutput object."""
        return f"AssignMask({self.output})"

    def modify_state(self, state: State) -> State:
        """Assign the state to the mask.

        Parameters
        ----------
        state : State
            The state dictionary.

        Returns
        -------
        State
            The masked state dictionary.
        """
        state = state.copy()

        state["latitudes"] = self._assign_mask(state["latitudes"])
        state["longitudes"] = self._assign_mask(state["longitudes"])
        for field in state["fields"]:
            state["fields"][field] = self._assign_mask(state["fields"][field])

        return state

    def _assign_mask(self, array: np.ndarray):
        shape = array.shape[:-1] + self.mask.shape
        res = np.full(shape, self.fill_value, dtype=array.dtype)
        if array.ndim == 1:
            res[self.mask] = array
        else:
            res[..., self.mask] = array
        return res



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import os
import threading
from pathlib import Path

import numpy as np

from anemoi.inference.context import Context
from anemoi.inference.types import ProcessorConfig
from anemoi.inference.types import State

from ..decorators import ensure_path
from ..decorators import main_argument
from ..output import Output
from . import output_registry

LOG = logging.getLogger(__name__)


# In case HDF5 was not compiled with thread safety on
LOCK = threading.RLock()


@output_registry.register("netcdf")
@main_argument("path")
@ensure_path("path")
class NetCDFOutput(Output):
    """NetCDF output class."""

    def __init__(
        self,
        context: Context,
        path: Path,
        variables: list[str] | None = None,
        post_processors: list[ProcessorConfig] | None = None,
        output_frequency: int | None = None,
        write_initial_state: bool | None = None,
        float_size: str = "f4",
        missing_value: float | None = np.nan,
    ) -> None:
        """Initialise the NetCDF output object.

        Parameters
        ----------
        context : dict
            The context dictionary.
        path : Path
            The path to save the NetCDF file to.
            If the parent directory does not exist, it will be created.
        post_processors : Optional[List[ProcessorConfig]], default None
            Post-processors to apply to the input
        output_frequency : int, optional
            The frequency of output, by default None.
        write_initial_state : bool, optional
            Whether to write the initial state, by default None.
        float_size : str, optional
            The size of the float, by default "f4".
        missing_value : float, optional
            The missing value, by default np.nan.
        """

        super().__init__(
            context,
            variables=variables,
            post_processors=post_processors,
            output_frequency=output_frequency,
            write_initial_state=write_initial_state,
        )

        from netCDF4 import Dataset

        self.path = path
        self.ncfile: Dataset | None = None
        self.float_size = float_size
        self.missing_value = missing_value
        if self.write_step_zero:
            self.extra_time = 1
        else:
            self.extra_time = 0

    def __repr__(self) -> str:
        """Return a string representation of the NetCDFOutput object."""
        return f"NetCDFOutput({self.path})"

    def open(self, state: State) -> None:
        """Open the NetCDF file and initialize dimensions and variables.

        Parameters
        ----------
        state : State
            The state dictionary.
        """
        from netCDF4 import Dataset

        with LOCK:
            if self.ncfile is not None:
                return

        # If the file exists, we may get a 'Permission denied' error
        if os.path.exists(self.path):
            os.remove(self.path)

        with LOCK:
            self.ncfile = Dataset(self.path, "w", format="NETCDF4")

        state = self.post_process(state)

        compression = {}  # dict(zlib=False, complevel=0)

        values = len(state["latitudes"])

        time = 0
        self.reference_date = state["date"]
        if (time_step := getattr(self.context, "time_step", None)) and (
            lead_time := getattr(self.context, "lead_time", None)
        ):
            time = lead_time // time_step
            time += self.extra_time

        if reference_date := getattr(self.context, "reference_date", None):
            self.reference_date = reference_date

        with LOCK:
            self.values_dim = self.ncfile.createDimension("values", values)
            self.time_dim = self.ncfile.createDimension("time", time)
            self.time_var = self.ncfile.createVariable("time", "i4", ("time",), **compression)

            self.time_var.units = f"seconds since {self.reference_date}"
            self.time_var.long_name = "time"
            self.time_var.calendar = "gregorian"

        with LOCK:
            latitudes = state["latitudes"]

            self.latitude_var = self.ncfile.createVariable("latitude", self.float_size, ("values",), **compression)
            self.latitude_var.units = "degrees_north"
            self.latitude_var.long_name = "latitude"

            longitudes = state["longitudes"]

            self.longitude_var = self.ncfile.createVariable("longitude", self.float_size, ("values",), **compression)
            self.longitude_var.units = "degrees_east"
            self.longitude_var.long_name = "longitude"

            self.latitude_var[:] = latitudes
            self.longitude_var[:] = longitudes

        self.vars = {}

        self.n = 0

    def ensure_variables(self, state: State) -> None:
        """Ensure that all variables are created in the NetCDF file.

        Parameters
        ----------
        state : State
            The state dictionary.
        """

        values = len(state["latitudes"])

        compression = {}  # dict(zlib=False, complevel=0)

        for name in state["fields"].keys():
            if self.skip_variable(name):
                continue

            if name in self.vars:
                continue

            chunksizes = (1, values)

            while np.prod(chunksizes) > 1000000:
                chunksizes = tuple(int(np.ceil(x / 2)) for x in chunksizes)

            with LOCK:
                missing_value = self.missing_value

                self.vars[name] = self.ncfile.createVariable(
                    name,
                    self.float_size,
                    ("time", "values"),
                    chunksizes=chunksizes,
                    fill_value=missing_value,
                    **compression,
                )

                self.vars[name].fill_value = missing_value
                self.vars[name].missing_value = missing_value

    def write_step(self, state: State) -> None:
        """Write the state.

        Parameters
        ----------
        state : State
            The state dictionary.
        """

        self.ensure_variables(state)

        step = state["date"] - self.reference_date
        self.time_var[self.n] = step.total_seconds()

        for name, value in state["fields"].items():
            if self.skip_variable(name):
                continue

            with LOCK:
                LOG.debug(f"🚧🚧🚧🚧🚧🚧 XXXXXX {name}, {self.n}, {value.shape}")
                self.vars[name][self.n] = value

        self.n += 1

    def close(self) -> None:
        """Close the NetCDF file."""
        if self.ncfile is not None:
            with LOCK:
                self.ncfile.close()
            self.ncfile = None



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest

from anemoi.inference.metadata import Metadata
from anemoi.inference.testing.mock_checkpoint import mock_load_metadata


@pytest.mark.parametrize(
    "initial, patch, expected",
    [
        (
            {"config": {"dataloader": {"dataset": "abc"}}},
            {"config": {"dataloader": {"something": {"else": "123"}}}},
            {"dataset": "abc", "something": {"else": "123"}},
        ),
        (
            {"config": {"dataloader": [{"dataset": "abc"}, {"dataset": "xyz"}]}},
            {"config": {"dataloader": {"cutout": [{"dataset": "123"}, {"dataset": "456"}]}}},
            {"cutout": [{"dataset": "123"}, {"dataset": "456"}]},
        ),
        (
            {"config": {"dataloader": "abc"}},
            {"config": {"dataloader": "xyz"}},
            "xyz",
        ),
    ],
)
def test_patch(initial, patch, expected):
    metadata = Metadata(initial)
    assert metadata._metadata == initial

    metadata.patch(patch)
    assert metadata._metadata["config"]["dataloader"] == expected


def test_constant_fields_patch():
    model_metadata = mock_load_metadata("unit/checkpoints/atmos.json", supporting_arrays=False)
    metadata = Metadata(model_metadata)

    fields = ["z", "sdor", "slor", "lsm"]
    metadata.patch({"dataset": {"constant_fields": fields}})
    assert metadata._metadata["dataset"]["constant_fields"] == fields

    # check that the rest of the metadata is still the same after patching
    metadata._metadata["dataset"].pop("constant_fields")
    assert model_metadata == metadata._metadata



# (C) Copyright 2025 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from anemoi.utils.registry import Registry

from anemoi.inference.context import Context
from anemoi.inference.processor import Processor
from anemoi.inference.types import ProcessorConfig

post_processor_registry = Registry(__name__)


def create_post_processor(context: Context, config: ProcessorConfig) -> Processor:
    """Create a post-processor.

    Parameters
    ----------
    context : Context
        The context for the post-processor.
    config : Configuration
        The configuration for the post-processor.

    Returns
    -------
    Processor
        The created post-processor.
    """
    return post_processor_registry.from_config(config, context)



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os
from copy import deepcopy
from typing import Any

import yaml

from anemoi.inference.testing import files_for_tests

# "2t", "10u", "10v", "msl", "lsm", "z", "tcc", "tp", "cos_latitudes", "insolation"
#   0,   1,     2,     3,     4,    5,    6,      7,        8,            9

SIMPLE_METADATA = {
    "config": {
        "training": {
            "multistep_input": 2,
            "precision": "16-mixed",
        },
        "data": {
            "timestep": "6h",
            "forcing": ["lsm", "z"],
            "diagnostic": ["tcc", "tp"],
        },
    },
    "data_indices": {
        "data": {
            "input": {"full": [0, 1, 2, 3, 4, 5, 8, 9], "forcing": [4, 5, 8, 9], "diagnostic": [6, 7]},
            "output": {"full": [0, 1, 2, 3, 6, 7]},
        },
        "model": {
            "input": {"full": [0, 1, 2, 3, 4, 5, 6, 7], "forcing": [4, 5, 6, 7], "prognostic": [0, 1, 2, 3]},
            "output": {"full": [0, 1, 2, 3, 4, 5], "prognostic": [0, 1, 2, 3]},
        },
    },
    "dataset": {
        "variables": ["2t", "10u", "10v", "msl", "lsm", "z", "tcc", "tp", "cos_latitude", "insolation"],
        "shape": [365, 10, 1, 40320],
        "variables_metadata": {
            "2t": {"mars": {"param": "2t", "levtype": "sfc"}},
            "10u": {"mars": {"param": "10u", "levtype": "sfc"}},
            "10v": {"mars": {"param": "10v", "levtype": "sfc"}},
            "msl": {"mars": {"param": "msl", "levtype": "sfc"}},
            "lsm": {"constant_in_time": True, "mars": {"param": "lsm", "levtype": "sfc"}},
            "z": {"constant_in_time": True, "mars": {"param": "z", "levtype": "sfc"}},
            "tcc": {"mars": {"param": "tcc", "levtype": "sfc"}},
            "tp": {"accumulated": True, "mars": {"param": "tp", "levtype": "sfc"}},
            "cos_latitude": {"computed_forcing": True, "constant_in_time": True},
            "insolation": {"computed_forcing": True, "constant_in_time": False},
        },
        "data_request": {"grid": "O96"},
    },
}


def mock_load_metadata(path: str | None, *, supporting_arrays: bool = True) -> tuple[dict[str, Any], dict[str, Any]]:
    """Load metadata from a YAML file.

    Parameters
    ----------
    path : str, optional
        The path to the checkpoint file.
    supporting_arrays : bool, optional
        Whether to include supporting arrays, by default True.

    Returns
    -------
    dict
        The loaded metadata.
    """

    if path is None:
        metadata = SIMPLE_METADATA
    else:
        if not os.path.isabs(path):
            path = files_for_tests(path)
        name, _ = os.path.splitext(path)
        for ext in (".yaml", ".json"):
            path = f"{name}{ext}"
            if os.path.exists(path):
                break

        with open(path) as f:
            metadata = yaml.safe_load(f)

    arrays: dict[str, Any] = {}

    if supporting_arrays:
        return metadata, arrays

    return metadata


def minimum_mock_checkpoint(metadata: dict[str, Any]) -> dict[str, Any]:
    """Create a minimum mock checkpoint from the given metadata.

    Parameters
    ----------
    metadata : dict
        The metadata to create the mock checkpoint from.

    Returns
    -------
    dict
        The minimum mock checkpoint.
    """

    KEEP = {"dataset.variables_metadata", "data_indices.data", "data_indices.model"}

    metadata = deepcopy(metadata)

    def drop(metadata: dict[str, Any], reference: dict[str, Any], *path: str) -> None:
        """Recursively drop keys from metadata that are not in the reference.

        Parameters
        ----------
        metadata : dict
            The metadata to modify.
        reference : dict
            The reference metadata to compare against.
        path : str
            The current path in the metadata.
        """

        if ".".join(path) in KEEP:
            return

        for k, v in list(metadata.items()):
            if k not in reference:
                del metadata[k]
            else:
                if isinstance(v, dict):
                    drop(metadata[k], reference[k], *path, k)

    drop(metadata, SIMPLE_METADATA)

    variables_metadata = metadata["dataset"]["variables_metadata"]

    for k, v in variables_metadata.items():
        mars = v.get("mars", {})
        for key in ("class", "date", "time", "hdate", "domain", "expver"):
            mars.pop(key, None)

    return metadata


def mock_torch_load(path: str, map_location: Any, weights_only: bool) -> Any:
    """Load a mock torch model for testing purposes.

    Parameters
    ----------
    path : str
        The path to the model file.
    map_location : Any
        The device on which to load the model.
    weights_only : bool
        Whether to load only the weights.

    Returns
    -------
    Any
        The mock torch model.
    """
    from .mock_model import MockModel

    assert weights_only is False, "Not implemented"

    metadata, arrays = mock_load_metadata(path)

    return MockModel(metadata, arrays)


class MockRunConfiguration:
    """Mock class for loading run configurations from `files_for_tests`."""

    @classmethod
    def load(cls, path: str, *args, **kwargs) -> dict[str, Any]:
        """Load a run configuration from a given path.

        Parameters
        ----------
        path : str
            The path to the run configuration file.
        *args : tuple
            Additional positional arguments.
        **kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        dict
            The loaded run configuration.
        """
        from anemoi.inference.config.run import RunConfiguration

        if not os.path.isabs(path):
            path = files_for_tests(path)

        return RunConfiguration.load(path, *args, **kwargs)



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import json
import logging
import re
from abc import abstractmethod
from typing import Any

from earthkit.data.utils.dates import to_datetime

from anemoi.inference.types import FloatArray
from anemoi.inference.types import ProcessorConfig
from anemoi.inference.types import State

from ..grib.encoding import grib_keys
from ..grib.templates.manager import TemplateManager
from ..output import Output

LOG = logging.getLogger(__name__)


class HindcastOutput:
    """Hindcast output class."""

    def __init__(self, reference_year: int) -> None:
        """Initialize the HindcastOutput object.

        Parameters
        ----------
        reference_year : int
            The reference year.
        """

        self.reference_year = reference_year

    def __call__(self, variable: Any, values: FloatArray, template: object, keys: dict) -> tuple:
        """Call the HindcastOutput object.

        Parameters
        ----------
        variable : Any
            The variable object.
        values : FloatArray
            The values array.
        template : object
            The template object.
        keys : dict
            The keys dictionary.

        Returns
        -------
        tuple
            The modified values, template, and keys.
        """

        if "date" not in keys:
            assert template.metadata("hdate", default=None) is None, template
            date = template.metadata("date")
        else:
            date = keys.pop("date")

        for k in ("date", "hdate", "eps", "productDefinitionTemplateNumber"):
            keys.pop(k, None)

        keys["localDefinitionNumber"] = 30
        keys["dataDate"] = int(to_datetime(date).strftime("%Y%m%d"))
        keys["referenceDate"] = int(to_datetime(date).replace(year=self.reference_year).strftime("%Y%m%d"))

        return values, template, keys


class Matching:

    def __init__(self, config):

        self.conditions = {}
        for k, v in config.items():
            self.conditions[re.compile(k)] = v

    def patch(self, variable, request):

        for k, v in self.conditions.items():
            m = k.match(variable.name)
            if m:
                patch = {}
                for kk, vv in v.items():
                    if isinstance(vv, str) and vv.startswith("\\"):
                        vv = m.group(int(vv[1:]))
                    patch[kk] = vv
                return patch

        return {}


class PatchOutput:
    def __init__(self, *patches):
        self.patches = []
        for patch in patches:
            assert isinstance(patch, dict), patch
            assert len(patch) == 1, patch
            key = list(patch.keys())[0]
            assert key == "variable", patch
        self.patches.append(Matching(patch[key]))

    def __call__(self, variable: Any, values: FloatArray, template: object, keys: dict) -> tuple:
        original_keys = keys.copy()
        result = keys.copy()
        for patch in self.patches:
            result.update(patch.patch(variable, original_keys))

        return values, template, result


MODIFIERS = dict(hindcast=HindcastOutput, patches=PatchOutput)


def modifier_factory(modifiers: list) -> list:
    """Create a list of modifier instances.

    Parameters
    ----------
    modifiers : list
        A list of modifier configurations.

    Returns
    -------
    list
        A list of modifier instances.
    """

    if modifiers is None:
        return []

    if not isinstance(modifiers, list):
        modifiers = [modifiers]

    result = []
    for modifier in modifiers:
        assert isinstance(modifier, dict), modifier
        assert len(modifier) == 1, modifier

        klass = list(modifier.keys())[0]
        if isinstance(modifier[klass], list):
            result.append(MODIFIERS[klass](*modifier[klass]))
        else:
            result.append(MODIFIERS[klass](**modifier[klass]))

    return result


class BaseGribOutput(Output):
    """Handles grib."""

    def __init__(
        self,
        context: dict,
        post_processors: list[ProcessorConfig] | None = None,
        *,
        encoding: dict[str, Any] | None = None,
        templates: list[str] | str | None = None,
        grib1_keys: dict[str, Any] | None = None,
        grib2_keys: dict[str, Any] | None = None,
        modifiers: list[str] | None = None,
        variables: list[str] | None = None,
        output_frequency: int | None = None,
        write_initial_state: bool | None = None,
    ) -> None:
        """Initialise the GribOutput object.

        Parameters
        ----------
        context : dict
            The context dictionary.
        post_processors : Optional[List[ProcessorConfig]] = None
            Post-processors to apply to the input
        encoding : dict, optional
            The encoding dictionary, by default None.
        templates : list or str, optional
            The templates list or string, by default None.
        grib1_keys : dict, optional
            The grib1 keys dictionary, by default None.
        grib2_keys : dict, optional
            The grib2 keys dictionary, by default None.
        modifiers : list, optional
            The list of modifiers, by default None.
        output_frequency : int, optional
            The frequency of output, by default None.
        write_initial_state : bool, optional
            Whether to write the initial state, by default None.
        variables : list, optional
            The list of variables, by default None.
        """

        super().__init__(
            context,
            variables=variables,
            post_processors=post_processors,
            output_frequency=output_frequency,
            write_initial_state=write_initial_state,
        )
        self._first = True

        self.encoding = encoding if encoding is not None else {}
        self.grib1_keys = grib1_keys if grib1_keys is not None else {}
        self.grib2_keys = grib2_keys if grib2_keys is not None else {}

        self.modifiers = modifier_factory(modifiers)
        self.variables = variables

        self.ensemble = False
        for d in (self.grib1_keys, self.grib2_keys, self.encoding):
            if "eps" in d:
                self.ensemble = d["eps"]
                break
            if d.get("type") in ("pf", "cf"):
                self.ensemble = True
                break

        self.template_manager = TemplateManager(self, templates)

        self.ensemble = False
        for d in (self.grib1_keys, self.grib2_keys, self.encoding):
            if "eps" in d:
                self.ensemble = d["eps"]
                break
            if d.get("type") in ("pf", "cf"):
                self.ensemble = True
                break

        self.template_manager = TemplateManager(self, templates)

    def write_initial_state(self, state: State) -> None:
        """Write the initial step of the state.

        Parameters
        ----------
        state : State
            The state object.
        """
        # We trust the GribInput class to provide the templates
        # matching the input state

        if not self.write_step_zero:
            return

        state = state.copy()

        self.reference_date = state["date"]
        state.setdefault("step", datetime.timedelta(0))

        for name in state["fields"].keys():
            if self.skip_variable(name):
                continue

            variable = self.typed_variables[name]

            if variable.is_computed_forcing:
                continue

            template = self.template(state, name)
            if template is None:
                # grib output only reliably works when we have grib input, everything else relies on external templates
                raise ValueError(
                    f"No grib template found for initial state param `{name}`. Try setting `write_initial_state` to `false`."
                )

        return self.write_step(self.post_process(state))

    def write_step(self, state: State) -> None:
        """Write a step of the state.

        Parameters
        ----------
        state : State
            The state object.
        """

        reference_date = self.reference_date or self.context.reference_date
        assert reference_date is not None, "No reference date set"

        step = state["step"]
        previous_step = state.get("previous_step")
        start_steps = state.get("start_steps", {})

        for name in state["fields"].keys():
            if self.skip_variable(name):
                continue

            values = state["fields"][name]
            keys = {}

            variable = self.typed_variables[name]

            if variable.is_computed_forcing:
                continue

            param = variable.grib_keys.get("param", name)

            template = self.template(state, name)

            keys.update(self.encoding)

            keys = grib_keys(
                values=values,
                template=template,
                date=int(reference_date.strftime("%Y%m%d")),
                time=reference_date.hour * 100,
                step=step,
                param=param,
                variable=variable,
                ensemble=self.ensemble,
                keys=keys,
                grib1_keys=self.grib1_keys,
                grib2_keys=self.grib2_keys,
                previous_step=previous_step,
                start_steps=start_steps,
            )

            for modifier in self.modifiers:
                values, template, keys = modifier(variable, values, template, keys)

            if LOG.isEnabledFor(logging.DEBUG):
                LOG.info("Encoding GRIB %s\n%s", template, json.dumps(keys, indent=4))

            try:
                self.write_message(values, template=template, **keys)
            except Exception:
                LOG.error("Error writing field %s", name)
                LOG.error("Template: %s", template)
                LOG.error("Keys:\n%s", json.dumps(keys, indent=4, default=str))
                raise

    @abstractmethod
    def write_message(self, message: FloatArray, *args: Any, **kwargs: Any) -> None:
        """Write a message to the grib file.

        Parameters
        ----------
        message : FloatArray
            The message array.
        *args : Any
            Additional arguments.
        **kwargs : Any
            Additional keyword arguments.
        """
        pass

    def template(self, state: State, name: str) -> object:
        """Get the template for a variable.

        Parameters
        ----------
        state : State
            The state object.
        name : str
            The variable name.

        Returns
        -------
        object
            The template object.
        """
        return self.template_manager.template(name, state, self.typed_variables)

    def template_lookup(self, name: str) -> dict:
        """Lookup the template for a variable.

        Parameters
        ----------
        name : str
            The variable name.

        Returns
        -------
        dict
            The template dictionary.
        """
        return self.encoding



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import warnings
from typing import Any

from anemoi.inference.config import Configuration
from anemoi.inference.context import Context
from anemoi.inference.types import ProcessorConfig
from anemoi.inference.types import State

from ..output import ForwardOutput

LOG = logging.getLogger(__name__)


class MaskedOutput(ForwardOutput):
    """Apply mask output class.

    Parameters
    ----------
    context : dict
        The context dictionary.
    mask : Any
        The mask.
    output : dict
        The output configuration dictionary.
    variables : list, optional
        The list of variables to extract, by default None.
    post_processors : Optional[List[ProcessorConfig]], default None
        Post-processors to apply to the input
    output_frequency : int, optional
        The frequency of output, by default None.
    write_initial_state : bool, optional
        Whether to write the initial state, by default None.
    """

    def __init__(
        self,
        context: Context,
        *,
        mask: Any,
        output: Configuration,
        variables: list[str] | None = None,
        post_processors: list[ProcessorConfig] | None = None,
        output_frequency: int | None = None,
        write_initial_state: bool | None = None,
    ) -> None:

        warnings.warn(
            f"The {self.__class__.__name__} is deprecated and will be removed in a future release. "
            "Use extraction post-processors instead.",
            DeprecationWarning,
        )
        super().__init__(
            context,
            output,
            variables=variables,
            post_processors=post_processors,
            output_frequency=output_frequency,
            write_initial_state=write_initial_state,
        )
        self.mask = mask

    def modify_state(self, state: State) -> State:
        """Apply the mask to the state.

        Parameters
        ----------
        state : State
            The state dictionary.

        Returns
        -------
        State
            The masked state dictionary.
        """
        state = state.copy()
        state["fields"] = state["fields"].copy()
        state["latitudes"] = state["latitudes"][self.mask]
        state["longitudes"] = state["longitudes"][self.mask]

        for field in state["fields"]:
            data = state["fields"][field]
            if data.ndim == 1:
                data = data[self.mask]
            else:
                data = data[..., self.mask]
            state["fields"][field] = data

        return state

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"{self.__class__.__name__}({self.mask}, {self.output})"



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import argparse
import json
import logging
import sys
from argparse import ArgumentParser
from argparse import Namespace
from collections.abc import Callable
from typing import Any

from anemoi.utils.dates import frequency_to_timedelta
from earthkit.data.utils.dates import to_datetime

from anemoi.inference.checkpoint import Checkpoint
from anemoi.inference.metadata import VARIABLE_CATEGORIES
from anemoi.inference.types import DataRequest
from anemoi.inference.types import Date
from anemoi.inference.variables import Variables

from ..config.run import RunConfiguration
from ..inputs.mars import postproc
from ..runners import create_runner
from . import Command

LOG = logging.getLogger(__name__)


def print_request(verb, request, file=sys.stdout):
    r = [verb]
    for k, v in request.items():
        if not isinstance(v, (list, tuple, set)):
            v = [v]
        v = [str(_) for _ in v]
        v = "/".join(v)
        r.append(f"{k}={v}")

    r = ",\n   ".join(r)
    print(r, file=file)
    print(file=file)


def checkpoint_to_requests(
    checkpoint: Checkpoint,
    *,
    date: Date,
    staging_dates: str | None = None,
    forecast_dates: bool = False,
    use_grib_paramid: bool = False,
    dont_fail_for_missing_paramid: bool = False,
    extra: list[str] | None = None,
    patch_request: Callable[[DataRequest], DataRequest] | None = None,
    use_scda: bool = False,
    include: list[str] | None = None,
    exclude: list[str] | None = None,
    lead_time: Any | None = None,
    time_step: Any | None = None,
) -> list[DataRequest]:
    """Convert a checkpoint to a list of data requests.

    Parameters
    ----------
    checkpoint : Checkpoint
        The checkpoint object containing the necessary data.
    date : Date
        The date for the data request.
    staging_dates : str, optional
        Path to a file with staging dates.
    forecast_dates : bool, optional
        Whether to use forecast dates (for forcings).
    use_grib_paramid : bool, optional
        Whether to use GRIB parameter IDs.
    dont_fail_for_missing_paramid : bool, optional
        Whether to ignore missing parameter IDs.
    extra : list of str, optional
        Additional request values.
    patch_request : Callable[[DataRequest], DataRequest], optional
        Function to patch the data request.
    use_scda : bool, optional
        Whether to use SCDA stream for 6/18 input time.
    include : Optional[List[str]]
        Categories to include.
    exclude : Optional[List[str]]
        Categories to exclude.
    lead_time : Optional[Any]
        The lead time for the data request (used with forecast_dates).
    time_step : Optional[Any]
        The time step for the data request (used with forecast_dates).

    Returns
    -------
    List[DataRequest]
        A list of data requests.
    """
    # TODO: Move this to the runner

    LOG.info("Converting checkpoint to requests")
    LOG.info("Include categories: %s", include)
    LOG.info("Exclude categories: %s", exclude)

    variables = checkpoint.select_variables(include=include, exclude=exclude)

    if not variables:
        return []

    area = checkpoint.area
    grid = checkpoint.grid

    more = postproc(grid, area)

    for r in extra or []:
        k, v = r.split("=")
        more[k] = v

    if staging_dates:
        dates = set()
        with open(staging_dates) as f:
            for line in f:
                date = to_datetime(line.strip())
                dates.update([date + h for h in checkpoint.lagged])
        dates = sorted(dates)
    else:
        date = to_datetime(date)
        dates = [date + h for h in checkpoint.lagged]

    if forecast_dates:
        new_dates = set()
        for date in dates:
            d = date
            last = d + lead_time
            while d <= last:
                new_dates.add(d)
                d += time_step

        dates = sorted(new_dates)

    requests = []
    for r in checkpoint.mars_requests(
        dates=dates,
        variables=variables,
        use_grib_paramid=use_grib_paramid,
        patch_request=patch_request,
        always_split_time=use_scda,
        dont_fail_for_missing_paramid=dont_fail_for_missing_paramid,
    ):
        r = r.copy()
        r.update(more)
        if use_scda:
            _patch_scda(r)
        requests.append(r)

    return requests


# Custom type function to parse and validate comma-separated input
def comma_separated_list(value):

    items = value.split(",")
    invalid = set()
    for item in items:
        for bit in item.split("+"):
            if bit not in VARIABLE_CATEGORIES:
                invalid.add(bit)
    if invalid:
        raise argparse.ArgumentTypeError(
            f"Invalid value(s): {', '.join(invalid)}. Allowed values are: {', '.join(VARIABLE_CATEGORIES)}"
        )
    return items


class RetrieveCmd(Command):
    """Used by prepml."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.description = self.__doc__
        command_parser.add_argument(
            "config",
            type=str,
            help="Path to config file. Can be omitted to pass config with overrides and defaults.",
        )
        command_parser.add_argument("--defaults", action="append", help="Sources of default values.")
        command_parser.add_argument("--date", type=str, help="Date")
        command_parser.add_argument("--output", type=str, default=None, help="Output file")
        command_parser.add_argument("--staging-dates", type=str, help="Path to a file with staging dates")
        command_parser.add_argument("--forecast-dates", action="store_true", help="Use forecast dates (for forcings)")
        command_parser.add_argument("--extra", action="append", help="Additional request values. Can be repeated")
        command_parser.add_argument("--use-scda", action="store_true", help="Use scda stream for 6/18 input time")
        command_parser.add_argument("--use-grib-paramid", action="store_true", help="Use paramId instead of param.")
        command_parser.add_argument(
            "--dont-fail-for-missing-paramid",
            action="store_true",
            help="Do not fail if a parameter ID is missing.",
        )
        command_parser.add_argument(
            "--include",
            type=comma_separated_list,
            help="Comma-separated list of variable categories to include",
        )
        command_parser.add_argument(
            "--exclude",
            type=comma_separated_list,
            help="Comma-separated list of variable categories to exclude",
        )

        # This is a alias to pairs of include/exclude
        command_parser.add_argument(
            "--input-type",
            default="default",
            choices=sorted(Variables.input_types()),
            help="Type of input variables to retrieve.",
        )

        command_parser.add_argument("--mars", action="store_true", help="Write requests for MARS retrieval")
        command_parser.add_argument(
            "--target", default="input.grib", help="Target path for the MARS retrieval requests"
        )
        command_parser.add_argument("--verb", default="retrieve", help="Verb for the MARS retrieval requests")
        command_parser.add_argument("overrides", nargs="*", help="Overrides as key=value")

    def run(self, args: Namespace) -> None:
        """Run the retrieve command.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """
        if "=" in args.config:
            args.overrides.append(args.config)
            args.config = {}

        config: RunConfiguration = RunConfiguration.load(args.config, args.overrides, defaults=args.defaults)

        runner = create_runner(config)
        lead_time = frequency_to_timedelta(config.lead_time)
        time_step = frequency_to_timedelta(runner.checkpoint.timestep)

        if args.staging_dates is None and args.date is None:
            raise ValueError("Either 'date' or 'staging_dates' must be provided.")

        if args.input_type is not None:
            include_exclude = Variables.input_type_to_include_exclude(args.input_type)
            if "include" in include_exclude:
                args.include = sorted(set(args.include or []) | set(include_exclude["include"]))
            if "exclude" in include_exclude:
                args.exclude = sorted(set(args.exclude or []) | set(include_exclude["exclude"]))

        requests = checkpoint_to_requests(
            runner.checkpoint,
            date=args.date,
            extra=args.extra,
            staging_dates=args.staging_dates,
            forecast_dates=args.forecast_dates,
            use_grib_paramid=config.use_grib_paramid or args.use_grib_paramid,
            dont_fail_for_missing_paramid=args.dont_fail_for_missing_paramid,
            patch_request=runner.patch_data_request,
            use_scda=args.use_scda,
            include=args.include if args.include else None,
            exclude=args.exclude if args.exclude else None,
            lead_time=lead_time,
            time_step=time_step,
        )

        if len(requests) > 1:
            requests[0]["target"] = args.target

        if args.output and args.output != "-":
            f = open(args.output, "w")
        else:
            f = sys.stdout

        if args.mars:
            # Write requests in MARS format
            for i, request in enumerate(requests):
                print_request(args.verb, request, file=f)

        else:
            json.dump(requests, f, indent=4)


def _patch_scda(request: dict[str, Any]) -> None:
    """Patch the SCDA stream in the request if necessary.
    ECMWF operational data has stream oper for 00 and 12 UTC and scda for 06 and 18 UTC.

    Parameters
    ----------
    request : dict
        The request dictionary to be patched.
    """

    if request.get("class", "od") != "od":
        return
    if request.get("type", "an") not in ("an", "fc"):
        return
    if request.get("stream", "oper") not in ("oper", "scda"):
        return

    request_time = int(request.get("time", 12))
    if request_time > 100:
        request_time = int(request_time / 100)

    if request_time in (6, 18):
        request["stream"] = "scda"


command = RetrieveCmd



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging

from anemoi.inference.config import Configuration
from anemoi.inference.context import Context
from anemoi.inference.types import ProcessorConfig

from . import output_registry
from .masked import MaskedOutput

LOG = logging.getLogger(__name__)


@output_registry.register("apply_mask")
class ApplyMaskOutput(MaskedOutput):
    """Apply mask output class."""

    def __init__(
        self,
        context: Context,
        *,
        mask: str,
        output: Configuration,
        variables: list[str] | None = None,
        post_processors: list[ProcessorConfig] | None = None,
        output_frequency: int | None = None,
        write_initial_state: bool | None = None,
    ) -> None:
        """Parameters
        ----------
        context : dict
            The context dictionary.
        mask : str
            The mask identifier.
        output : dict
            The output configuration dictionary.
        post_processors : Optional[List[ProcessorConfig]], default None
            Post-processors to apply to the input
        output_frequency : int, optional
            The frequency of output, by default None.
        write_initial_state : bool, optional
            Whether to write the initial state, by default None.
        """
        super().__init__(
            context,
            mask=context.checkpoint.load_supporting_array(mask),
            output=output,
            variables=variables,
            post_processors=post_processors,
            output_frequency=output_frequency,
            write_initial_state=write_initial_state,
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from __future__ import annotations

import logging
import os
from datetime import datetime
from typing import Any
from typing import TypeVar

from earthkit.data.utils.dates import to_datetime
from omegaconf import DictConfig
from omegaconf import ListConfig
from omegaconf import OmegaConf
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import field_validator

LOG = logging.getLogger(__name__)

T = TypeVar("T", bound="Configuration")


class Configuration(BaseModel):
    """Configuration class."""

    model_config = ConfigDict(extra="forbid")

    date: datetime | None = None
    """The starting date for the forecast. If not provided, the date will depend on the selected Input object. If a string, it is parsed by :func:`earthkit.data.utils.dates`."""

    @field_validator("date", mode="before")
    @classmethod
    def to_datetime(cls, date: str | int | datetime | None) -> datetime | None:
        if date is not None:
            return to_datetime(date)

    @classmethod
    def load(
        cls: type[T],
        path: str | dict[str, Any],
        overrides: list[str] | list[dict] | str | dict = [],
        defaults: str | list[str] | dict | None = None,
    ) -> T:
        """Load the configuration.

        Parameters
        ----------
        path : Union[str, dict]
            Path to the configuration file or a dictionary containing the configuration.
        overrides : Union[List[str], List[dict], str, dict], optional
            List of overrides to apply to the configuration, by default [].
        defaults : Optional[Union[str, List[str], dict]], optional
            Default values to set in the configuration, by default None.

        Returns
        -------
        Configuration
            The loaded configuration.
        """

        configs: list[DictConfig | ListConfig] = []

        # Set default values
        if defaults is not None:
            if not isinstance(defaults, list):
                defaults = [defaults]
            for d in defaults:
                if isinstance(d, str):
                    configs.append(OmegaConf.load(d))
                    continue
                configs.append(OmegaConf.create(d))

        # Load the user configuration
        if isinstance(path, dict):
            configs.append(OmegaConf.create(path))
        else:
            configs.append(OmegaConf.load(path))

        if not isinstance(overrides, list):
            overrides = [overrides]

        # unsafe merge should be fine as we don't re-use the original configs
        oc_config = OmegaConf.unsafe_merge(*configs)

        for override in overrides:
            if isinstance(override, dict):
                oc_config = OmegaConf.unsafe_merge(oc_config, OmegaConf.create(override))
            else:
                # use from_dotlist to use OmegaConf split
                # which allows for "param.val" or "param[val]".
                override_conf = OmegaConf.from_dotlist([override])
                # We can't directly merge reconstructed with the config because
                # omegaconf prefers parsing digits (like 0 in key.0) into dict keys
                # rather than lists.
                # Instead, we provide a reference config and we try to merge the override
                # into the reference and keep types provided by the reference.
                oc_config = OmegaConf.unsafe_merge(_merge_configs(oc_config, override_conf))

        resolved_config = OmegaConf.to_container(oc_config, resolve=True)

        config = cls.model_validate(resolved_config)

        # Set environment variables found in the configuration
        # as soon as possible
        for key, value in config.env.items():
            os.environ[key] = str(value)

        return config


def _merge_configs(ref_conf: Any, new_conf: Any) -> Any:
    """Recursively merges a new OmegaConf object into a reference OmegaConf object

    Parameters
    ----------
    ref_conf : Any
        reference OmegaConf object. Should be a DictConfig or ListConfig.
    new_conf : Any
        new OmegaConf object.

    Returns
    -------
    Any
        The merged OmegaConf config
    """
    if isinstance(new_conf, DictConfig) and len(new_conf):
        key, rest = next(iter(new_conf.items()))
        key = str(key)
    elif isinstance(new_conf, ListConfig) and len(new_conf):
        key, rest = 0, new_conf[0]
    else:
        return new_conf
    if isinstance(ref_conf, ListConfig):
        if isinstance(key, str) and not key.isdigit():
            raise ValueError(f"Expected int key, got {key}")
        index = int(key)
        if index < len(ref_conf):
            LOG.debug(f"key {key} is used as list key in list{ref_conf}")
            ref_conf[index] = _merge_configs(ref_conf[index], rest)
        elif index == len(ref_conf):
            LOG.debug(f"key {key} is used to append to list {ref_conf}")
            ref_conf.append(rest)
        else:
            raise IndexError(f"key {key} out of range for list {ref_conf} of length {len(ref_conf)}")
        return ref_conf
    elif isinstance(ref_conf, DictConfig) and key in ref_conf:
        ref_conf[key] = _merge_configs(ref_conf[key], rest)
        return ref_conf
    elif isinstance(ref_conf, DictConfig):
        ref_conf[key] = rest
        return ref_conf
    else:
        raise ValueError(f"ref is of unexpected type {type(ref_conf)}. Should be ListConfig or DictConfig")



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import json
import logging
from functools import cached_property
from typing import Any

import numpy as np

from anemoi.inference.context import Context
from anemoi.inference.types import Date
from anemoi.inference.types import FloatArray
from anemoi.inference.types import ProcessorConfig
from anemoi.inference.types import State

from ..input import Input
from . import input_registry

LOG = logging.getLogger(__name__)


class DatasetInput(Input):
    """Handles `anemoi-datasets` dataset as input."""

    def __init__(
        self,
        context: Context,
        *,
        open_dataset_args: tuple[Any, ...],
        open_dataset_kwargs: dict[str, Any],
        grid_indices: Any = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the DatasetInput.

        Parameters
        ----------
        context : Any
            The context in which the input is used.
        open_dataset_args : Tuple[Any, ...]
            Arguments for the dataset.
        open_dataset_kwargs : Dict[str, Any]
            Keyword arguments for the dataset.
        grid_indices : Optional[Any]
            Indices to reduce the input grid. If None, the full grid is used.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, **kwargs)

        self.open_dataset_args = open_dataset_args
        self.open_dataset_kwargs = open_dataset_kwargs

        if context.verbosity > 0:
            LOG.info(
                "Opening dataset with\nargs=%s\nkwargs=%s",
                json.dumps(open_dataset_args, indent=4),
                json.dumps(
                    open_dataset_kwargs,
                    indent=4,
                ),
            )

        if grid_indices is None and "grid_indices" in context.checkpoint._supporting_arrays:
            grid_indices = context.checkpoint.load_supporting_array("grid_indices")
            if context.verbosity > 0:
                LOG.info(
                    "Loading supporting array `grid_indices` from checkpoint, \
                    the input grid will be reduced accordingly."
                )

        self.grid_indices = slice(None) if grid_indices is None else grid_indices

    @cached_property
    def ds(self) -> Any:
        """Return the dataset."""
        from anemoi.datasets import open_dataset

        dataset = open_dataset(*self.open_dataset_args, **self.open_dataset_kwargs)
        if self.variables is not None:
            dataset = open_dataset(dataset, select=self.variables)

        return dataset

    @cached_property
    def latitudes(self) -> FloatArray:
        """Return the latitudes."""
        return self.ds.latitudes

    @cached_property
    def longitudes(self) -> FloatArray:
        """Return the longitudes."""
        return self.ds.longitudes

    def __repr__(self) -> str:
        """Return a string representation of the DatasetInput."""
        return f"DatasetInput({self.open_dataset_args}, {self.open_dataset_kwargs})"

    def create_input_state(self, *, date: Date | None = None) -> State:
        """Create the input state for the given date.

        Parameters
        ----------
        date : Optional[Any]
            The date for which to create the input state.

        Returns
        -------
        Dict[str, Any]
            The created input state.
        """
        if date is None:
            raise ValueError("`date` must be provided")

        latitudes = self.ds.latitudes
        longitudes = self.ds.longitudes

        input_state = dict(
            date=date,
            latitudes=latitudes[self.grid_indices],
            longitudes=longitudes[self.grid_indices],
            fields=dict(),
        )

        fields = input_state["fields"]

        date = np.datetime64(date)
        dates = [date + np.timedelta64(h) for h in self.checkpoint.lagged]

        data = self._load_dates(dates)

        if data.shape[2] != 1:
            raise ValueError(f"Ensemble data not supported, got {data.shape[2]} members")

        requested_variables = set(self.input_variables())
        for i, variable in enumerate(self.ds.variables):
            if variable not in requested_variables:
                continue
            # Squeeze the data to remove the ensemble dimension
            values = np.squeeze(data[:, i], axis=1)
            fields[variable] = values[:, self.grid_indices]

            if self.context.trace:
                self.context.trace.from_input(variable, self)

        return input_state

    def load_forcings_state(self, *, dates: list[Date], current_state: State) -> State:
        """Load the forcings state for the given variables and dates.

        Parameters
        ----------
        dates : List[Any]
            List of dates for which to load the forcings.
        current_state : State
            The current state of the input.

        Returns
        -------
        State
            The loaded forcings state.
        """
        data = self._load_dates(dates)  # (date, variables, ensemble, values)

        requested_variables = np.array([self.ds.name_to_index[v] for v in self.variables])
        data = data[:, requested_variables]
        # Squeeze the data to remove the ensemble dimension
        data = np.squeeze(data, axis=2)
        # Reorder the dimensions to (variable, date, values)
        data = np.swapaxes(data, 0, 1)

        # apply reduction to `grid_indices`
        data = data[..., self.grid_indices]

        fields = {v: data[i] for i, v in enumerate(self.variables)}

        return dict(
            fields=fields,
            dates=dates,
            latitudes=self.latitudes,
            longitudes=self.longitudes,
        )

    def _load_dates(self, dates: list[Date]) -> Any:
        """Load the data for the given dates.

        Parameters
        ----------
        dates : List[Any]
            List of dates for which to load the data.

        Returns
        -------
        Any
            The loaded data.
        """
        # TODO: use the fact that the dates are sorted

        dataset_dates = self.ds.dates

        idx = []
        for d in dates:
            (i,) = np.where(dataset_dates == d)
            if len(i) == 0:
                raise ValueError(
                    f"Date {d} not found in the dataset, available dates: {dataset_dates[0]}...{dataset_dates[-1]} by {self.ds.frequency}"
                )
            assert len(i) == 1, f"Multiple dates found for {d}"
            idx.append(int(i[0]))

        if len(idx) == 1:
            s = slice(idx[0], idx[0] + 1)
        else:
            diff = idx[1] - idx[0]
            if not all(i == diff for i in np.diff(idx)):
                # TODO: remove that restriction
                raise ValueError("Dates do not have the same frequency")
            s = slice(idx[0], idx[-1] + 1, diff)

        return self.ds[s]


@input_registry.register("dataset")
class DatasetInputArgsKwargs(DatasetInput):
    """Handles `anemoi-datasets` dataset as input."""

    trace_name = "dataset/provided"

    def __init__(
        self,
        context: Context,
        *args: Any,
        use_original_paths: bool = False,
        variables: list[str] | None,
        pre_processors: list[ProcessorConfig] | None = None,
        grid_indices=None,
        purpose: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the DatasetInputArgsKwargs.

        Parameters
        ----------
        context : Context
            The context in which the input is used.
        use_original_paths : bool
            Whether to use original paths.
        """
        if not args and not kwargs:
            args, kwargs = context.checkpoint.open_dataset_args_kwargs(use_original_paths=use_original_paths)

            # TODO: remove start/end from the arguments

            LOG.warning("No arguments provided to open_dataset, using the default arguments:")

            cmd = "open_dataset("
            for arg in args:
                cmd += f"{arg}, "
            for key, value in kwargs.items():
                cmd += f"{key}={value}, "

            LOG.warning("%s", cmd)

        super().__init__(
            context,
            variables=variables,
            pre_processors=pre_processors,
            grid_indices=grid_indices,
            open_dataset_args=args,
            open_dataset_kwargs=kwargs,
            purpose=purpose,
        )


class DataloaderInput(DatasetInput):
    """Handles `anemoi-datasets` dataset as input."""

    def __init__(
        self,
        context: Context,
        *,
        use_original_paths: bool = False,
        **kwargs: Any,
    ) -> None:
        """Initialize the DataloaderInput.

        Parameters
        ----------
        context : Any
            The context in which the input is used.
        name : str
            The name of the dataloader.
        use_original_paths : bool
            Whether to use original paths.
        """
        open_dataset_args, open_dataset_kwargs = context.checkpoint.open_dataset_args_kwargs(
            use_original_paths=use_original_paths,
            from_dataloader=self.name,
        )

        super().__init__(
            context,
            open_dataset_args=open_dataset_args,
            open_dataset_kwargs=open_dataset_kwargs,
            **kwargs,
        )


@input_registry.register("test")
class TestInput(DataloaderInput):
    """Handles `anemoi-datasets` dataset as input."""

    trace_name = "dataset/test"
    name = "test"


@input_registry.register("validation")
class ValidationInput(DataloaderInput):
    """Handles `anemoi-datasets` dataset as input."""

    trace_name = "dataset/validation"
    name = "validation"


@input_registry.register("training")
class TrainingInput(DataloaderInput):
    """Handles `anemoi-datasets` dataset as input."""

    trace_name = "dataset/training"
    name = "training"



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging

import earthkit.data as ekd

from anemoi.inference.context import Context
from anemoi.inference.types import Date
from anemoi.inference.types import State

from . import input_registry
from .grib import GribInput

LOG = logging.getLogger(__name__)


@input_registry.register("icon_grib_file")
class IconInput(GribInput):
    """Handles grib files from ICON."""

    # TODO: this code will become a plugin in the future.

    trace_name = "icon file"

    def __init__(
        self,
        context: Context,
        *,
        path: str,
        grid: str,
        refinement_level_c: int,
        **kwargs,
    ) -> None:
        """Initialize the IconInput.

        Parameters
        ----------
        context : Any
            The context in which the input is used.
        path : str
            The path to the ICON grib file.
        grid : str
            The grid type.
        refinement_level_c : int
            The refinement level.
        pre_processors : Optional[List[ProcessorConfig]], default None
            Pre-processors to apply to the input
        namer : Optional[Any]
            Optional namer for the input.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, **kwargs)
        self.path = path
        self.grid = grid
        self.refinement_level_c = refinement_level_c

    def create_input_state(self, *, date: Date | None) -> State:
        """Creates the input state for the given date.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The created input state.
        """
        from anemoi.transform.grids.icon import icon_grid

        latitudes, longitudes = icon_grid(self.grid, self.refinement_level_c)

        return self._create_input_state(
            ekd.from_source("file", self.path),
            variables=None,
            date=date,
            latitudes=latitudes,
            longitudes=longitudes,
        )

    def load_forcings_state(self, *, dates: list[Date], current_state: State) -> State:
        """Loads the forcings state for the given variables and dates.

        Parameters
        ----------
        dates : List[Date]
            List of dates for which to load the forcings state.
        current_state : State
            The current state.

        Returns
        -------
        Any
            The loaded forcings state.
        """
        return self._load_forcings_state(
            ekd.from_source("file", self.path),
            dates=dates,
            current_state=current_state,
        )



import sys

import pytest

from anemoi.inference.grib.encoding import render_template


@pytest.mark.skipif(
    sys.version_info < (3, 10),
    reason="Format specifier requires python 3.10 or higher.",
)
@pytest.mark.parametrize(
    "template, handle, expected",
    [
        (
            "{dateTime}_{step:03}.grib",
            {"dateTime": "202001011200", "step": 6},
            "202001011200_006.grib",
        ),
        (
            "{validityDate}{validityTime:04}.grib",
            {"validityDate": "20200101", "validityTime": 900},
            "202001010900.grib",
        ),
    ],
)
def test_render_template(template, handle, expected):
    """Test the render_template function.

    NOTE: we mock the handle with a dictionary.
    """
    result = render_template(template, handle)
    assert result == expected



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import itertools
import logging
import warnings
from typing import TYPE_CHECKING
from typing import Any

from anemoi.utils.config import DotDict
from anemoi.utils.dates import frequency_to_timedelta as to_timedelta
from pydantic import BaseModel

from anemoi.inference.config import Configuration
from anemoi.inference.input import Input
from anemoi.inference.output import Output
from anemoi.inference.processor import Processor
from anemoi.inference.types import IntArray

from ..forcings import BoundaryForcings
from ..forcings import ComputedForcings
from ..forcings import ConstantForcings
from ..forcings import CoupledForcings
from ..forcings import Forcings
from ..inputs import create_input
from ..outputs import create_output
from ..post_processors import create_post_processor
from ..pre_processors import create_pre_processor
from ..runner import Runner
from . import runner_registry

LOG = logging.getLogger(__name__)

if TYPE_CHECKING:
    import torch


@runner_registry.register("default")
class DefaultRunner(Runner):
    """Default runner class for inference.

    This class provides the default implementation for running inference.
    """

    def __init__(self, config: Configuration) -> None:
        """Initialize the DefaultRunner.

        Parameters
        ----------
        config : Configuration
            Configuration for the runner.
        """

        if isinstance(config, dict):
            # So we get the dot notation
            config = DotDict(config)

        # Remove that when the Pydantic model is ready
        if isinstance(config, BaseModel):
            config = DotDict(config.model_dump())

        self.config = config

        super().__init__(
            config.checkpoint,
            device=config.device,
            precision=config.precision,
            allow_nans=config.allow_nans,
            verbosity=config.verbosity,
            use_grib_paramid=config.use_grib_paramid,
            patch_metadata=config.patch_metadata,
            development_hacks=config.development_hacks,
            output_frequency=config.output_frequency,
            write_initial_state=config.write_initial_state,
            initial_state_categories=config.initial_state_categories,
            trace_path=config.trace_path,
            use_profiler=config.use_profiler,
            typed_variables=config.typed_variables,
        )

    def predict_step(
        self, model: "torch.nn.Module", input_tensor_torch: "torch.Tensor", **kwargs: Any
    ) -> "torch.Tensor":
        for key, value in self.config.predict_kwargs.items():
            if key in kwargs:
                warnings.warn(
                    f"`predict_kwargs` contains illegal kwarg `{key}`. This kwarg is set by the runner and will be ignored."
                )
                continue
            kwargs[key] = value

        return super().predict_step(model, input_tensor_torch, **kwargs)

    def execute(self) -> None:
        """Execute the runner."""

        if self.config.description is not None:
            LOG.info("%s", self.config.description)

        lead_time = to_timedelta(self.config.lead_time)

        # This may be used by Output objects to compute the step
        self.lead_time = lead_time
        self.time_step = self.checkpoint.timestep

        output = self.create_output()

        # In case the constant forcings are from another input, combine them here
        # So that they are in considered in the `write_initial_state`

        prognostic_input = self.create_prognostics_input()
        LOG.info(f"📥 Prognostic input: {prognostic_input}")
        prognostic_state = prognostic_input.create_input_state(date=self.config.date)
        self._check_state(prognostic_state, "prognostics")

        constants_input = self.create_constant_coupled_forcings_input()
        LOG.info(f"📥 Constant forcings input: {constants_input}")
        constants_state = constants_input.create_input_state(date=self.config.date)
        self._check_state(constants_state, "constant_forcings")

        forcings_input = self.create_dynamic_forcings_input()
        LOG.info(f"📥 Dynamic forcings input: {forcings_input}")
        forcings_state = forcings_input.create_input_state(date=self.config.date)
        self._check_state(forcings_state, "dynamic_forcings")

        input_state = self._combine_states(
            prognostic_state,
            constants_state,
            forcings_state,
        )

        # This hook is needed for the coupled runner
        self.input_state_hook(constants_state)

        # For step-zero only
        initial_state = Output.reduce(
            self._initial_state(
                prognostic_state,
                constants_state,
                forcings_state,
            )
        )
        # Top-level post-processors on the other hand are applied on State and are executed here.
        LOG.info("Top-level post-processors: %s", self.post_processors)

        for processor in self.post_processors:
            initial_state = processor.process(initial_state)

        output.open(initial_state)

        LOG.info("write_initial_state: %s", output)
        output.write_initial_state(initial_state)

        for state in self.run(input_state=input_state, lead_time=lead_time):
            # Apply top-level post-processors
            for processor in self.post_processors:
                state = processor.process(state)
            output.write_state(state)

        output.close()

        if "accumulate_from_start_of_forecast" not in self.config.post_processors:
            LOG.warning(
                """
                🚧 The default accumulation behaviour has changed. 🚧
                🚧 Accumulation fields have NOT been accumulated from the beginning of the forecast. 🚧
                🚧 To accumulate from the beginning, set `post_processors: [accumulate_from_start_of_forecast]` 🚧
                """  # ecmwf/anemoi-inference#131
            )

    def create_output(self) -> Output:
        """Create the output.

        Returns
        -------
        Output
            The created output.
        """
        output = create_output(self, self.config.output)
        LOG.info("Output: %s", output)
        return output

    def create_constant_computed_forcings(self, variables: list[str], mask: IntArray) -> list[Forcings]:

        result = ComputedForcings(self, variables, mask)
        LOG.info("Constant computed forcing: %s", result)
        return [result]

    def create_dynamic_computed_forcings(self, variables: list[str], mask: IntArray) -> list[Forcings]:

        result = ComputedForcings(self, variables, mask)
        LOG.info("Dynamic computed forcing: %s", result)
        return [result]

    def _input_forcings(self, *names: str) -> dict[str, Any]:
        """Get the input forcings configuration.

        Parameters
        ----------
        names : list
            The name of the forcings configuration.

        Returns
        -------
        dict
            The input forcings configuration.
        """

        for name in names:
            if name.startswith("-"):
                deprecated = True
                name = name[1:]  # Remove the leading dash
            else:
                deprecated = False
            if self.config.get(name):
                if deprecated:
                    LOG.warning(
                        f"🚫 The `{name}` input forcings configuration is deprecated. "
                        f"Please use the `{names[0]}` configuration instead."
                    )
                if name != names[0]:
                    LOG.info(f"Loading `config.{names[0]}` from `config.{name}`")

                return self.config[name]

        return self.config.input

    #########################################################################################################
    def create_prognostics_input(self) -> Input:
        """Create the prognostics input.

        Returns
        -------
        Input
            The created prognostics input.
        """
        variables = self.variables.retrieved_prognostic_variables()
        config = self._input_forcings("prognostic_input", "input") if variables else "empty"
        input = create_input(self, config, variables=variables, purpose="prognostics")
        LOG.info("Prognostic input: %s", input)
        return input

    def create_constant_coupled_forcings_input(self) -> Input:
        """Create the constant coupled forcings input.

        Returns
        -------
        Input
            The created constant coupled forcings input.
        """
        variables = self.variables.retrieved_constant_forcings_variables()
        config = self._input_forcings("constant_forcings", "forcings", "input") if variables else "empty"
        input = create_input(self, config, variables=variables, purpose="constant_forcings")
        LOG.info("Constant coupled forcings input: %s", input)
        return input

    def create_dynamic_forcings_input(self) -> Input:
        """Create the dynamic forcings input.

        Returns
        -------
        Input
            The created dynamic forcings input.
        """
        variables = self.variables.retrieved_dynamic_forcings_variables()
        config = self._input_forcings("dynamic_forcings", "-forcings", "input") if variables else "empty"
        input = create_input(self, config, variables=variables, purpose="dynamic_forcings")
        LOG.info("Dynamic forcings input: %s", input)
        return input

    def create_boundary_forcings_input(self) -> Input:
        """Create the boundary forcings input.

        Returns
        -------
        Input
            The created boundary forcings input.
        """
        variables = self.variables.retrieved_boundary_forcings_variables()
        config = self._input_forcings("boundary_forcings", "-boundary", "forcings", "input") if variables else "empty"
        input = create_input(self, config, variables=variables, purpose="boundary_forcings")
        LOG.info("Boundary forcings input: %s", input)
        return input

    #########################################################################################################
    def create_constant_coupled_forcings(self, variables: list[str], mask: IntArray) -> list[Forcings]:
        """Create constant coupled forcings.

        Parameters
        ----------
        variables : list
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List[Forcings]
            The created constant coupled forcings.
        """
        input = self.create_constant_coupled_forcings_input()
        result = ConstantForcings(self, input, variables, mask)
        LOG.info("Constant coupled forcing: %s", result)

        return [result]

    def create_dynamic_coupled_forcings(self, variables: list[str], mask: IntArray) -> list[Forcings]:
        """Create dynamic coupled forcings.

        Parameters
        ----------
        variables : list
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List[Forcings]
            The created dynamic coupled forcings.
        """
        input = self.create_dynamic_forcings_input()
        result = CoupledForcings(self, input, variables, mask)
        LOG.info("Dynamic coupled forcing: %s", result)
        return [result]

    def create_boundary_forcings(self, variables: list[str], mask: IntArray) -> list[Forcings]:
        """Create boundary forcings.

        Parameters
        ----------
        variables : list
            The variables for the forcings.
        mask : IntArray
            The mask for the forcings.

        Returns
        -------
        List[Forcings]
            The created boundary forcings.
        """
        input = self.create_boundary_forcings_input()
        result = BoundaryForcings(self, input, variables, mask)
        LOG.info("Boundary forcing: %s", result)
        return [result]

    def create_pre_processors(self) -> list[Processor]:
        """Create pre-processors.

        Returns
        -------
        List[Processor]
            The created pre-processors.
        """
        result = []
        for processor in self.config.pre_processors:
            result.append(create_pre_processor(self, processor))

        LOG.info("Pre processors: %s", result)
        return result

    def create_post_processors(self) -> list[Processor]:
        """Create post-processors.

        Returns
        -------
        List[Processor]
            The created post-processors.
        """
        result = []
        for processor in self.config.post_processors:
            result.append(create_post_processor(self, processor))

        LOG.info("Post processors: %s", result)
        return result

    def _combine_states(self, *states: dict[str, Any]) -> dict[str, Any]:
        """Combine multiple states into one.

        Parameters
        ----------
        states : list
            The states to combine.

        Returns
        -------
        dict
            The combined state.
        """
        import numpy as np

        combined = states[0].copy()
        combined["fields"] = combined["fields"].copy()
        shape = None
        first_input = combined.get("_input")

        for state in states[1:]:

            this_input = state.get("_input")

            for name, values in itertools.chain(combined["fields"].items(), state.get("fields", {}).items()):
                if shape is None:
                    shape = values.shape
                elif shape != values.shape:
                    raise ValueError(
                        f"Field '{name}' has different shape in the states: "
                        f"{shape} and {values.shape}."
                        f" Input: {first_input} vs {this_input}."
                    )

            if not set(combined["fields"]).isdisjoint(state["fields"]):
                raise ValueError(
                    f"Some states have overlapping fields:"
                    f" {set(combined['fields']).intersection(state['fields'])}"
                    f" Input: {first_input} vs {this_input}."
                )

            combined["fields"].update(state.get("fields", {}))
            for key, value in state.items():
                if key == "fields":
                    continue

                if key.startswith("_"):
                    continue

                if combined.get(key) is None:
                    combined[key] = value
                    continue

                if value is None:
                    continue

                if type(combined[key]) is not type(value):
                    raise ValueError(
                        f"Key '{key}' has different types in the states: " f"{type(combined[key])} and {type(value)}."
                    )

                if isinstance(value, np.ndarray) and isinstance(combined[key], np.ndarray):
                    if not np.array_equal(combined[key], value):
                        raise ValueError(
                            f"Key '{key}' has different array values in the states: "
                            f"{combined[key]} and {value}."
                            f" Input: {first_input} vs {this_input}."
                        )
                    continue

                if combined[key] != value:
                    raise ValueError(
                        f"Key '{key}' has different values in the states: "
                        f"{combined[key]} and {value} ({shape})."
                        f" Input: {first_input} vs {this_input}."
                    )

        return combined

    def _initial_state(
        self, prognostic_state: dict[str, Any], constants_state: dict[str, Any], forcings_state: dict[str, Any]
    ) -> dict[str, Any]:
        """Create the initial state for the output.

        Parameters
        ----------
        prognostic_state : dict
            The prognostic state.
        constants_state : dict
            The constant forcings state.
        forcings_state : dict
            The dynamic forcings state.

        Returns
        -------
        dict
            The initial state for the output.
        """
        states = []

        if "prognostics" in self.config.initial_state_categories:
            states.append(prognostic_state)

        if "constant_forcings" in self.config.initial_state_categories:
            states.append(constants_state)

        if "dynamic_forcings" in self.config.initial_state_categories:
            states.append(forcings_state)

        return self._combine_states(*states)

    def _check_state(self, state: dict[str, Any], title: str) -> None:
        """Check the state for consistency.

        Parameters
        ----------
        state : dict
            The state to check.
        title : str
            The title of the state (for logging).

        Raises
        ------
        ValueError
            If the state is not consistent.
        """

        if not isinstance(state, dict):
            raise ValueError(f"State '{title}' is not a dictionary: {state}")

        input = state.get("_input")

        if "fields" not in state:
            raise ValueError(f"State '{title}' does not contain 'fields': {state} ({input=})")

        shape = None

        for field, values in state["fields"].items():
            if shape is None:
                shape = values.shape
            elif shape != values.shape:
                raise ValueError(
                    f"Field '{field}' in state '{title}' has different shape: "
                    f"{shape} and {values.shape} ({input=})."
                )

        date = state.get("date")
        if date is None and len(state["fields"]) > 0:
            # date can be None for an empty input
            if not isinstance(date, datetime.datetime):
                raise ValueError(f"State '{title}' does not contain 'date', or it is not a datetime: {date} ({input=})")



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
from pathlib import Path
from typing import TYPE_CHECKING

from . import testing_registry

if TYPE_CHECKING:
    from anemoi.transform.variables import Variable

    from anemoi.inference.checkpoint import Checkpoint

LOG = logging.getLogger(__name__)

# checks here are used in the integration tests and can be reused in other test suites
# checks must use keyword arguments only and must have a **kwargs


@testing_registry.register("check_grib")
def check_grib(
    *,
    file: Path,
    expected_variables: list["Variable"],
    grib_keys: dict = {},
    check_accum: str | None = None,
    check_nans=False,
    **kwargs,
) -> None:
    LOG.info(f"Checking GRIB file: {file}")
    import earthkit.data as ekd
    import numpy as np

    ds = ekd.from_source("file", file)

    assert len(ds) > 0, "No fields found in the GRIB file."

    params = set(ds.metadata("param"))
    expected_params = [var.param for var in expected_variables]
    assert (
        set(expected_params) == params
    ), f"Expected parameters {set(expected_params)} do not match found parameters {params}."

    for field in ds:
        for key, value in grib_keys.items():
            assert field[key] == value, f"Key {key} does not match expected value {value}."
        assert not np.all(field.values == 0), f"Field {field} is zero."

        if check_nans:
            assert not any(np.isnan(field.values)), f"Field {field} contains NaN values."

    # check time continuity
    fields = ds.sel(param=expected_params[0])
    previous_field = fields[0]
    for field in fields[1:]:
        assert field["step"] > previous_field["step"] and field["valid_time"] > previous_field["valid_time"], (
            f"Field step {field['step']} (valid_time {field['valid_time']}) is not greater than previous field "
            f"step {previous_field['step']} (valid_time {previous_field['valid_time']})."
        )
        previous_field = field

    if not check_accum:
        return

    # check that the accumulation field is accumulating
    fields = list(ds.sel(param=check_accum))
    if len(fields) < 2:
        raise ValueError(f"No fields found for accumulation check: {check_accum}")

    averages = [np.average(field.values) for field in fields]
    assert all(curr > prev for prev, curr in zip(averages, averages[1:])), f"{check_accum} is not accumulating"


@testing_registry.register("check_with_xarray")
def check_with_xarray(
    *, file: Path, expected_variables: list["Variable"], check_accum: str | None = None, check_nans=False, **kwargs
) -> None:
    LOG.info(f"Checking file: {file}")
    import numpy as np
    import xarray as xr

    ds = xr.open_dataset(file)

    assert len(ds.data_vars) > 0, "No data found in the xarray compatible file."

    # lat and lon are variables in the file, but they are not variables of the model so skip them
    skip = {"latitude", "longitude"}
    params = set(ds.data_vars) - skip
    expected_params = [var.name for var in expected_variables]
    assert (
        set(expected_params) == params
    ), f"Expected parameters {set(expected_params)} do not match found parameters {params}."

    for var in ds.data_vars:
        if var in skip:
            continue
        assert not np.all(ds[var].values == 0), f"Variable {var} is zero."

        if check_nans:
            assert not np.isnan(ds[var].values).any(), f"Variable {var} contains NaN values."

    if not check_accum:
        return

    data_vars = list(ds.data_vars[check_accum])
    if len(data_vars) < 2:
        raise ValueError(f"No variables found for accumulation check: {check_accum}")

    averages = [np.average(data.values) for data in data_vars]
    assert all(curr > prev for prev, curr in zip(averages, averages[1:])), f"{check_accum} is not accumulating"


@testing_registry.register("check_lam_with_xarray")
def check_lam(
    *,
    file: Path,
    checkpoint: "Checkpoint",
    mask="lam_0",
    reference_date: str = None,
    reference_dataset={},
    reference_file=None,
    **kwargs,
) -> None:
    LOG.info(f"Checking LAM: {file}")
    import numpy as np
    import xarray as xr

    ds = xr.open_dataset(file)

    # check shape of inner region against the mask in the checkpoint
    mask = checkpoint.load_supporting_array(f"{mask}/cutout_mask")
    for var in ds.data_vars:
        assert ds[var].shape[-1] == np.sum(
            mask
        ), f"Variable {var} shape {ds[var].shape[-1]} does not match mask size {np.sum(mask)}."

    # check that the extracted inner region matches the reference input dataset
    # the mock model passes input to output, values in the output file should match the input dataset at the reference date
    if reference_dataset:
        from anemoi.datasets import open_dataset

        ref_ds = open_dataset(**reference_dataset, start=reference_date, end=reference_date)

        for var in checkpoint.prognostic_variables:
            assert var in ds.data_vars, f"Variable {var} not found in output file."
            ref_idx = ref_ds.name_to_index[var]
            for data in ds[var]:
                assert np.allclose(
                    data.values, ref_ds[0, ref_idx, 0, :]
                ), f"Variable {var} in output does not match reference data at {reference_date}."
    elif reference_file:
        # check against a reference file, implement when needed
        raise NotImplementedError("Reference file check is not implemented yet.")


@testing_registry.register("check_file_exist")
def check_file_exist(*, file: Path, **kwargs) -> None:
    LOG.info(f"Checking file exists: {file}")
    assert file.exists(), f"File {file} does not exist."
    assert file.stat().st_size > 0, f"File {file} is empty."


@testing_registry.register("check_files_in_directory")
def check_files_in_directory(*, file: Path, expected_files: int | None = None, **kwargs) -> None:
    LOG.info(f"Checking directory: {file}")
    assert file.exists() and file.is_dir(), f"Directory {file} does not exist or is not a directory."
    if expected_files is not None:
        actual_files = [f for f in file.iterdir() if f.is_file()]
        if expected_files < 0:
            assert len(actual_files) > 0, "Expected at least one file, but found none."
        else:
            assert (
                len(actual_files) == expected_files
            ), f"Expected {expected_files} files, but found {len(actual_files)}."



# (C) Copyright 2025- Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime
import functools
import json
import os
from collections.abc import Callable
from pathlib import Path
from typing import Any

from anemoi.utils.checkpoints import save_metadata
from anemoi.utils.registry import Registry

testing_registry = Registry(__name__)


def fake_checkpoints(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to mock checkpoints for testing.

    Parameters
    ----------
    func : Callable
        The function to be decorated.

    Returns
    -------
    Callable
        The decorated function.
    """

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        from unittest.mock import MagicMock
        from unittest.mock import patch

        from .mock_checkpoint import MockRunConfiguration
        from .mock_checkpoint import mock_load_metadata
        from .mock_checkpoint import mock_torch_load

        with (
            patch("anemoi.inference.checkpoint.load_metadata", mock_load_metadata),
            patch("anemoi.inference.provenance.validate_environment", MagicMock()),
            patch("torch.load", mock_torch_load),
            patch("anemoi.inference.metadata.USE_LEGACY", True),
            patch("anemoi.inference.tasks.runner.RunConfiguration", MockRunConfiguration),
        ):
            return func(*args, **kwargs)

    return wrapper


def save_fake_checkpoint(metadata: dict | str | Path, save_path: Path, supporting_arrays: dict = {}) -> None:
    """Create a fake PyTorch checkpoint.

    Parameters
    ----------
    metadata : dict, str, Path
        Metadata dictionary or path to a json file.
    save_path : Path
        The path to the checkpoint file.
    supporting_arrays : dict
        Supporting arrays to be saved in the checkpoint.
    """
    import torch

    from anemoi.inference.testing.mock_model import SimpleMockModel

    if not isinstance(metadata, dict):
        with open(metadata) as f:
            metadata = json.load(f)

    model = SimpleMockModel(metadata, supporting_arrays)

    torch.save(model, save_path)

    save_metadata(
        save_path,
        metadata,
        supporting_arrays=supporting_arrays,
    )


def float_hash(s: str, date: datetime.datetime, accuracy: int = 1_000_000) -> float:
    """Hash a string and date to a float.

    Parameters
    ----------
    s : str
        The string to be hashed.
    date : datetime.datetime
        The date to be hashed.
    accuracy : int, optional
        The accuracy of the hash, by default 1_000_000

    Returns
    -------
    float
        The resulting hash as a float.
    """
    h = s + date.isoformat()
    return float(int.from_bytes(h.encode(), "little") % accuracy) / accuracy


def files_for_tests(name: str) -> str:
    """Get the path to the testing files.

    Parameters
    ----------
    name : str
        The name of the test file.

    Returns
    -------
    str
        The path to the test file.
    """
    # Running in GitHub Actions, the package was installed

    if "GITHUB_WORKSPACE" in os.environ:
        return os.path.join(os.environ["GITHUB_WORKSPACE"], "tests", name)

    # We assume that the test data is in the same directory as the test
    # and this is a development environment

    import anemoi.inference

    bits = os.path.normpath(anemoi.inference.__file__).split(os.path.sep)
    while len(bits) > 3 and (bits[-3], bits[-2], bits[-1]) != ("src", "anemoi", "inference"):
        bits.pop()

    for i in range(3):
        bits.pop()

    bits.append("tests")
    bits.append(name)
    return os.path.sep.join(bits)


class TestingContext:
    """A context for testing plugins."""

    pass



import argparse
import json
import re
import sys
import tempfile
import warnings
from pathlib import Path

import numpy as np
import yaml
from anemoi.utils.checkpoints import load_metadata
from anemoi.utils.remote import transfer
from anemoi.utils.remote.s3 import _list_objects
from anemoi.utils.sanitise import sanitise

warnings.filterwarnings(
    "ignore",
    message=".*DotDict.*",  # ignore DotDict immutable warning triggered by `transfer`
    category=UserWarning,
    module="anemoi.utils.config",
)

INTEGRATION_ROOT = Path(__file__).resolve().parent
S3_ROOT = "s3://ml-tests/test-data/samples/anemoi-integration-tests/inference"

parser = argparse.ArgumentParser(
    description="Add a new model for integration tests. Running this script will create a new model directory in tests/integration with metadata and config files, and upload necessary files to S3."
)
parser.add_argument(
    "model",
    type=str,
    help="Name of the model. Can only contain alphanumeric, underscores, hyphens, or dots.",
)
parser.add_argument(
    "checkpoint",
    type=Path,
    help="Path to the inference checkpoint file.",
)
parser.add_argument(
    "--files",
    "-f",
    type=Path,
    nargs="*",
    default=[],
    help="Additional files to upload to the model directory on S3.",
)
parser.add_argument(
    "--overwrite",
    "-o",
    action="store_true",
    help="Overwrite existing files.",
)
args = parser.parse_args()

model_path = INTEGRATION_ROOT / args.model
metadata_path = model_path / "metadata.json"
config_path = model_path / "config.yaml"

if not re.match(r"^[\w\-\.]+$", args.model):
    print("Model name must be a valid path name (alphanumeric, underscores, hyphens, or dots only).")
    sys.exit(1)

if not args.checkpoint.exists():
    print(f"Checkpoint file {args.checkpoint} does not exist.")
    sys.exit(1)

for file in args.files:
    if not file.exists():
        print(f"{file} does not exist.")
        sys.exit(1)

if model_path.exists() and not args.overwrite:
    print(f"Model directory `{args.model}` already exists in `{INTEGRATION_ROOT}`. Use --overwrite to replace it.")
    sys.exit(1)

model_path.mkdir(parents=True, exist_ok=True)

print(f"🚀 Adding model {args.model}...")
metadata, supporting_arrays = load_metadata(args.checkpoint, supporting_arrays=True)
metadata = sanitise(metadata, level=2)

# delete unused data
del metadata["tracker"]
del metadata["training"]
del metadata["model"]
metadata["config"] = {
    "data": metadata["config"]["data"],
    "training": metadata["config"]["training"],
}

# save metadata and example config file in the local repo
with open(metadata_path, "w") as f:
    json.dump(metadata, f, indent=2)

with open(config_path, "w") as f:
    yaml.dump(
        [
            {
                "name": "example-integration-test",
                "input": "input.grib",
                "output": "output.grib",
                "checks": [],
                "inference_config": {
                    "checkpoint": "${checkpoint:}",
                    "input": {"grib": "${input:}"},
                    "output": {"grib": "${output:}"},
                },
            }
        ],
        f,
        sort_keys=False,
    )

# upload files to S3
for file in args.files:
    s3_path = f"{S3_ROOT}/{args.model}/{file.name}"
    transfer(str(file), s3_path, overwrite=args.overwrite, resume=not args.overwrite)

with tempfile.TemporaryDirectory() as temp_dir:
    temp_dir_path = Path(temp_dir)
    for name, array in supporting_arrays.items():
        array_path = temp_dir_path / f"{name}.npy"
        array_path.parent.mkdir(parents=True, exist_ok=True)
        np.save(array_path, array)
        s3_path = f"{S3_ROOT}/{args.model}/supporting-arrays/{name}.npy"
        transfer(
            str(array_path),
            s3_path,
            overwrite=args.overwrite,
            resume=not args.overwrite,
        )

print("Done. Summary:")
print(f"💾 Files to be comitted, created in {model_path}:")
for item in model_path.iterdir():
    print(f" - {item.name}")

print(f"☁️ Files in {S3_ROOT}/{args.model}:")
for item in _list_objects(f"{S3_ROOT}/{args.model}"):
    print(f" - {item['Key'].split(f'{args.model}/')[-1]}")

print(f"✅ Model {args.model} added successfully.")



# (C) Copyright 2025 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import datetime
import logging
import sys
from typing import Any

import numpy as np

from anemoi.inference.types import BoolArray
from anemoi.inference.types import FloatArray

LOG = logging.getLogger(__name__)


class RolloutSource:
    """Represents a source of data that is a rollout."""

    trace_name = "rollout"


class UnknownSource:
    """Represents a source of data that is unknown."""

    trace_name = "?"


class UnchangedSource:
    """Represents a source of data that is unchanged."""

    trace_name = "unchanged"


class InputSource:
    """Represents a source of data that is an input."""

    def __init__(self, input: Any) -> None:
        self.input = input
        self.trace_name = input.trace_name


class Trace:
    """Implementation of a trace."""

    def __init__(self, path: bool | str) -> None:
        self.path = path
        self.file = sys.stdout if path is True else open(path, "w")

        self.sources: dict[str, Any] = {}
        self.extra: dict[str, Any] = {}

        print("-+" * 80, file=self.file)
        print("Trace:", datetime.datetime.now(), file=self.file)
        print("-+" * 80, file=self.file)
        print(file=self.file, flush=True)

    def write_input_tensor(
        self,
        date: datetime.datetime,
        fcstep: int,
        input_tensor: FloatArray,
        variable_to_input_tensor_index: dict[str, int],
        timestep: datetime.timedelta,
    ) -> None:
        """Write the input tensor details to the trace file.

        Parameters
        ----------
        date : datetime.datetime
            The date associated with the input tensor.
        fcstep : int
            The step number.
        input_tensor : FloatArray
            The input tensor.
        variable_to_input_tensor_index : Dict[str, int]
            Mapping from variable names to input tensor indices.
        timestep : datetime.timedelta
            The timestep.
        """
        print(f"Input tensor to forecast date {date-timestep} => {date}", input_tensor.shape, file=self.file)
        names = {v: k for k, v in variable_to_input_tensor_index.items()}
        assert len(input_tensor.shape) == 4
        assert input_tensor.shape[0] == 1
        lines = []

        for s in range(input_tensor.shape[1]):
            for i in range(input_tensor.shape[-1]):
                values = input_tensor[0, s, :, i]
                if s == 0:
                    lines.append(
                        [names[i], np.nanmin(values), np.nanmax(values), np.nanmean(values), np.nanstd(values)]
                    )
                else:
                    constant = np.all(values == input_tensor[0, 0, :, i])
                    constant = "C" if constant else ""
                    lines[i].extend(
                        [np.nanmin(values), np.nanmax(values), np.nanmean(values), np.nanstd(values), constant]
                    )

        unknown = UnknownSource()
        for line in lines:
            line.append(self.sources.get(line[0], unknown).trace_name)

        lines.insert(0, ["--------", "---", "---", "----", "---", "---", "---", "----", "---", "-", "---"])
        lines.insert(0, ["Variable", "Min", "Max", "Mean", "Std", "Min", "Max", "Mean", "Std", "C", "Source"])
        self.table(lines)

        print(file=self.file, flush=True)

    def write_output_tensor(
        self,
        date: datetime.datetime,
        fcstep: int,
        output_tensor: FloatArray,
        output_tensor_index_to_variable: dict[int, str],
        timestep: datetime.timedelta,
    ) -> None:
        """Write the output tensor details to the trace file.

        Parameters
        ----------
        date : datetime.datetime
            The date associated with the output tensor.
        fcstep : int
            The step number.
        output_tensor : FloatArray
            The output tensor.
        output_tensor_index_to_variable : Dict[int, str]
            Mapping from output tensor indices to variable names.
        timestep : datetime.timedelta
            The timestep.
        """
        print(f"Output tensor for {date}:", output_tensor.shape, file=self.file)
        assert len(output_tensor.shape) == 2
        names = output_tensor_index_to_variable
        lines = []
        for i in range(output_tensor.shape[-1]):
            values = output_tensor[:, i]
            lines.append([names[i], np.nanmin(values), np.nanmax(values), np.nanmean(values), np.nanstd(values)])

        lines.insert(0, ["--------", "---", "---", "----", "---"])
        lines.insert(0, ["Variable", "Min", "Max", "Mean", "Std"])
        self.table(lines)
        print(file=self.file, flush=True)

    def table(self, lines: list[list[Any]]) -> None:
        """Print a formatted table to the trace file.

        Parameters
        ----------
        lines : list[list[Any]]
            The table data to print.
        """
        print(file=self.file)

        def _(x: Any) -> str:
            if isinstance(x, float):
                return f"{x:g}"
            return str(x)

        lines = [[_(x) for x in line] for line in lines]

        lengths = [max(len(str(x)) for x in col) for col in zip(*lines)]
        for line in lines:
            print(" ".join(f"{x:{lengths[i]}}" for i, x in enumerate(line)), file=self.file)

    def from_source(self, name: str, source: Any, extra: Any = None) -> None:
        """Add a source to the trace.

        Parameters
        ----------
        name : str
            The name of the source.
        source : Any
            The source object.
        extra : Any, optional
            Additional information about the source.
        """
        if name in self.sources:
            old = self.sources[name]
            LOG.warning(
                f"Warning: source {name} already defined as '{old.trace_name}', changing to '{source.trace_name}'"
            )

        # assert name not in self.sources, name
        self.sources[name] = source
        self.extra[name] = extra

    def from_rollout(self, name: str) -> None:
        """Add a rollout source to the trace.

        Parameters
        ----------
        name : str
            The name of the rollout source.
        """
        self.from_source(name, RolloutSource())

    def from_input(self, name: str, input: Any) -> None:
        """Add an input source to the trace.

        Parameters
        ----------
        name : str
            The name of the input source.
        input : Any
            The input object.
        """
        self.from_source(name, InputSource(input))

    def reset_sources(self, reset: BoolArray, variable_to_input_tensor_index: dict[str, int]) -> None:
        """Reset the sources in the trace.

        Parameters
        ----------
        reset : list[bool]
            List indicating which sources to reset.
        variable_to_input_tensor_index : Dict[str, int]
            Mapping from variable names to input tensor indices.
        """
        self.sources = {}
        self.extra = {}
        names = {v: k for k, v in variable_to_input_tensor_index.items()}
        for i, on in enumerate(reset):
            if on:
                self.sources[names[i]] = UnchangedSource()



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import json
import logging
import os
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import NamedTuple

import numpy as np
import pytest
from anemoi.transform.variables.variables import VariableFromMarsVocabulary
from anemoi.utils.testing import TEST_DATA_URL
from anemoi.utils.testing import GetTestData
from omegaconf import ListConfig
from omegaconf import OmegaConf

from anemoi.inference.config.run import RunConfiguration
from anemoi.inference.runners import create_runner
from anemoi.inference.testing import save_fake_checkpoint
from anemoi.inference.testing import testing_registry

logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)

INTEGRATION_ROOT = Path(__file__).resolve().parent

# each model has its own folder in the integration tests directory
# and contains at least a metadata.json and config.yaml file
MODELS = [
    path.name
    for path in INTEGRATION_ROOT.iterdir()
    if path.is_dir() and (path / "metadata.json").exists() and (path / "config.yaml").exists()
]

# each model can have more than one test configuration, defined as a listconfig in config.yaml
# the integration test is parameterised over the models and their test configurations
MODEL_CONFIGS = [
    (model, config) for model in MODELS for config in OmegaConf.load(INTEGRATION_ROOT / model / "config.yaml")
]


class Setup(NamedTuple):
    config: OmegaConf
    output: Path


@pytest.fixture(params=MODEL_CONFIGS, ids=[f"{model}/{config.name}" for model, config in MODEL_CONFIGS])
def test_setup(request, get_test_data: GetTestData, tmp_path: Path) -> Setup:
    model, config = request.param
    input = config.input
    output = tmp_path / config.output
    inference_config = config.inference_config
    s3_path = f"anemoi-integration-tests/inference/{model}"

    # download input file(s)
    if not input:
        input = []
    if not isinstance(input, (list, ListConfig)):
        input = [input]
    input_data = [get_test_data(f"{s3_path}/{file}") for file in input]

    # change working directory to the input temporary directory, so the config could use relative paths to input files
    if input_data:
        workdir = Path(input_data[0]).parent
    else:
        workdir = tmp_path
    LOG.info(f"Changing working directory to {workdir}")
    os.chdir(workdir)

    # prepare checkpoint
    checkpoint_path = tmp_path / Path("checkpoint.ckpt")
    with open(INTEGRATION_ROOT / model / "metadata.json") as f:
        metadata = json.load(f)

    supporting_arrays = {}
    if array_names := list(metadata.get("supporting_arrays_paths", {})):

        def load_array(name):
            return name, np.load(get_test_data(f"{s3_path}/supporting-arrays/{name}.npy"))

        with ThreadPoolExecutor(max_workers=4) as executor:
            results = list(executor.map(load_array, array_names))

        supporting_arrays = dict(results)
        LOG.info(f"Supporting arrays: {supporting_arrays.keys()}")

    save_fake_checkpoint(metadata, checkpoint_path, supporting_arrays=supporting_arrays)

    # substitute inference config with real paths
    OmegaConf.register_new_resolver("input", lambda i=0: str(input_data[i]), replace=True)
    OmegaConf.register_new_resolver("output", lambda: str(output), replace=True)
    OmegaConf.register_new_resolver("checkpoint", lambda: str(checkpoint_path), replace=True)
    OmegaConf.register_new_resolver("s3", lambda: str(f"{TEST_DATA_URL}{s3_path}"), replace=True)

    # save the inference config to disk
    inference_config = OmegaConf.to_yaml(inference_config, resolve=True)
    LOG.info(f"Resolved config:\n{inference_config}")

    with open(tmp_path / "integration_test.yaml", "w") as f:
        f.write(inference_config)

    return Setup(config=config, output=output)


def test_integration(test_setup: Setup, tmp_path: Path) -> None:
    """Run the integration test suite."""

    overrides = {"lead_time": "48h", "device": "cpu"}
    LOG.info(f"Config overrides: {overrides}")

    config = RunConfiguration.load(
        tmp_path / "integration_test.yaml",
        overrides=overrides,
    )
    runner = create_runner(config)
    runner.execute()

    assert (test_setup.output).exists(), "Output file was not created."

    checkpoint_output_variables = _typed_variables_output(runner._checkpoint)
    LOG.info(f"Checkpoint output variables: {checkpoint_output_variables}")

    # run the checks defined in the test configuration
    # checks is a list of dicts, each with a single key-value pair
    for checks in test_setup.config.checks:
        check, kwargs = next(iter(checks.items()))
        # config can optionally pass expected output variables, by default it uses the checkpoint variables
        expected_variables_config = kwargs.pop("expected_variables", [])
        expected_variables = [
            VariableFromMarsVocabulary(var, {"param": var}) for var in expected_variables_config
        ] or checkpoint_output_variables

        testing_registry.create(
            check,
            file=test_setup.output,
            expected_variables=expected_variables,
            checkpoint=runner._checkpoint,
            **kwargs,
        )


def _typed_variables_output(checkpoint):
    output_variables = checkpoint.output_tensor_index_to_variable.values()
    return [checkpoint.typed_variables[name] for name in output_variables]



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import logging
from typing import Any

import earthkit.data as ekd
from earthkit.data.utils.dates import to_datetime

from anemoi.inference.context import Context
from anemoi.inference.types import DataRequest
from anemoi.inference.types import Date
from anemoi.inference.types import ProcessorConfig
from anemoi.inference.types import State

from . import input_registry
from .grib import GribInput
from .mars import postproc

LOG = logging.getLogger(__name__)


def retrieve(
    requests: list[DataRequest],
    grid: str | list[float] | None,
    area: list[float] | None,
    dataset: str | dict[str, Any],
    **kwargs: Any,
) -> ekd.FieldList:
    """Retrieve data from CDS.

    Parameters
    ----------
    requests : List[Dict[str, Any]]
        List of request dictionaries.
    grid : Optional[Union[str, List[float]]]
        Grid specification.
    area : Optional[List[float]]
        Area specification.
    dataset : Union[str, Dict[str, Any]]
        Dataset to use.
    **kwargs : Any
        Additional keyword arguments.

    Returns
    -------
    Any
        Retrieved data.
    """

    def _(r: DataRequest) -> str:
        mars = r.copy()
        for k, v in r.items():
            if isinstance(v, (list, tuple)):
                mars[k] = "/".join(str(x) for x in v)
            else:
                mars[k] = str(v)

        return ",".join(f"{k}={v}" for k, v in mars.items())

    pproc = postproc(grid, area)

    result = ekd.from_source("empty")
    for r in requests:
        if isinstance(dataset, str):
            d = dataset
        elif isinstance(dataset, dict):
            # Get dataset from intersection of keys between request and dataset dict
            search_dataset = dataset.copy()
            while isinstance(search_dataset, dict):
                keys = set(r.keys()).intersection(set(search_dataset.keys()))
                if len(keys) == 0:
                    raise KeyError(
                        f"While searching for dataset, could not find any valid key in dictionary: {r.keys()}, {search_dataset}"
                    )
                key = list(keys)[0]
                if r[key] not in search_dataset[key]:
                    if "*" in search_dataset[key]:
                        search_dataset = search_dataset[key]["*"]
                        continue

                    raise KeyError(
                        f"Dataset dictionary does not contain key {r[key]!r} in {key!r}: {dict(search_dataset[key])}."
                    )
                search_dataset = search_dataset[key][r[key]]

            d = search_dataset

        r.update(pproc)
        r.update(kwargs)

        LOG.debug("%s", _(r))

        result += ekd.from_source("cds", d, r)

    return result


@input_registry.register("cds")
class CDSInput(GribInput):
    """Get input fields from CDS."""

    trace_name = "cds"

    def __init__(
        self,
        context: Context,
        pre_processors: list[ProcessorConfig] | None = None,
        *,
        dataset: str | dict[str, Any],
        namer: Any | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the CDSInput.

        Parameters
        ----------
        context : Context
            The context in which the input is used.
        pre_processors : Optional[List[ProcessorConfig]], default None
            Pre-processors to apply to the input
        dataset : Union[str, Dict[str, Any]]
            The dataset to use.
        namer : Optional[Any]
            Optional namer for the input.
        **kwargs : Any
            Additional keyword arguments.
        """
        super().__init__(context, pre_processors, namer=namer)

        self.dataset = dataset
        self.kwargs = kwargs

    def create_input_state(self, *, date: Date | None) -> State:
        """Create the input state for the given date.

        Parameters
        ----------
        date : Optional[Date]
            The date for which to create the input state.

        Returns
        -------
        State
            The created input state.
        """
        if date is None:
            date = to_datetime(-1)
            LOG.warning("CDSInput: `date` parameter not provided, using yesterday's date: %s", date)

        return self._create_input_state(
            self.retrieve(
                self.variables,
                [date + h for h in self.checkpoint.lagged],
            ),
            variables=self.variables,
            date=date,
        )

    def retrieve(self, variables: list[str], dates: list[Date]) -> Any:
        """Retrieve data for the given variables and dates.

        Parameters
        ----------
        variables : List[str]
            List of variables to retrieve.
        dates : List[Date]
            List of dates for which to retrieve data.

        Returns
        -------
        Any
            Retrieved data.
        """

        requests = self.checkpoint.mars_requests(
            variables=variables,
            dates=dates,
            use_grib_paramid=self.context.use_grib_paramid,
            patch_request=self.patch_data_request,
        )

        if not requests:
            raise ValueError(f"No requests for {variables} ({dates})")

        return retrieve(
            requests, self.checkpoint.grid, self.checkpoint.area, dataset=self.dataset, expver="0001", **self.kwargs
        )

    def load_forcings_state(self, *, dates: list[Date], current_state: State) -> State:
        """Load the forcings state for the given variables and dates.

        Parameters
        ----------
        dates : List[Date]
            The list of dates for which to load the forcings state.
        current_state : State
            The current state to be updated with the loaded forcings state.

        Returns
        -------
        Any
            The loaded forcings state.
        """
        return self._load_forcings_state(
            self.retrieve(self.variables, dates),
            dates=dates,
            current_state=current_state,
        )



# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import json
import logging
import sys
import warnings
from argparse import ArgumentParser
from argparse import Namespace
from collections.abc import Callable
from typing import Any

import rich
import yaml

from ..checkpoint import Checkpoint
from . import Command

CORE = ("models", "training", "graphs")


SKIP = {"anemoi-training"}

SKIP |= set(sys.stdlib_module_names) | set(sys.builtin_module_names)

PACKAGES = {
    "sklearn": "scikit-learn",
    "attr": "attrs",
    "google-protobuf": "protobuf",
}


class InspectCmd(Command):
    """Inspect the contents of a checkpoint file."""

    def add_arguments(self, command_parser: ArgumentParser) -> None:
        """Add arguments to the command parser.

        Parameters
        ----------
        command_parser : ArgumentParser
            The argument parser to which the arguments will be added.
        """
        command_parser.add_argument("path", help="Path to the checkpoint.")
        group = command_parser.add_mutually_exclusive_group(required=True)
        group.add_argument("--variables", action="store_true", help="List the training variables and their categories.")
        group.add_argument(
            "--requirements",
            action="store_true",
            help="Print a Python's requirements.txt based on the versions of the packages used during training.",
        )
        group.add_argument(
            "--datasets", action="store_true", help="Print the arguments passed to anemoi-dataset during training."
        )

        group.add_argument("--dump", action="store_true", help="Dump information from the checkpoint.")

        command_parser.add_argument("--json", action="store_true", help="Output in JSON format (with dump option)")
        command_parser.add_argument("--validate", action="store_true", help="Validate the environment.")

    def run(self, args: Namespace) -> None:
        """Run the inspect command.

        Parameters
        ----------
        args : Namespace
            The arguments passed to the command.
        """

        c = Checkpoint(args.path)

        if args.validate:
            c.validate_environment()
            return

        if args.variables:
            self.variables(c, args)
            return

        if args.requirements:
            self.requirements(c, args)
            return

        if args.datasets:
            self.datasets(c, args)
            return

        if args.dump:
            self.dump(c, args)
            return

    def dump(self, c: Checkpoint, args: Namespace) -> None:

        if args.json:
            # turn off all other logging so json output can be piped cleanly
            logging.disable(logging.INFO)
            warnings.filterwarnings("ignore")

        def _(f: Callable[[], Any]) -> Any:
            """Wrapper function to handle exceptions.

            Parameters
            ----------
            f : Callable
                The function to be called.

            Returns
            -------
            Any
                The result of the function call or the exception message.
            """
            try:
                return f()
            except Exception as e:
                return str(e)

        data = {}
        for name in sorted(dir(c)):

            if name.startswith("_"):
                continue

            data[name] = _(lambda: getattr(c, name))

        if args.json:
            print(json.dumps(data, indent=None, default=str))
        else:
            for key, value in data.items():
                print(key, ":")
                print("  ", json.dumps(value, indent=4, default=str))
                print()

    def variables(self, c: Checkpoint, args: Namespace) -> None:
        """Print the variable categories in the checkpoint.

        Parameters
        ----------
        c : Checkpoint
            The checkpoint object.
        args : Namespace
            The command-line arguments.
        """
        c.print_variable_categories(print=rich.print)

    def indices(self, c: Checkpoint, args: Namespace) -> None:
        """Print the variable indices in the checkpoint.

        Parameters
        ----------
        c : Checkpoint
            The checkpoint object.
        args : Namespace
            The command-line arguments.
        """
        c.print_indices(print=rich.print)

    def datasets(self, c: Checkpoint, args: Namespace) -> None:
        """Print the dataset arguments and keyword arguments for opening datasets.

        Parameters
        ----------
        c : Checkpoint
            The checkpoint object.
        args : Namespace
            The command-line arguments.
        """
        open_dataset_args, open_dataset_kwargs = c.open_dataset_args_kwargs(use_original_paths=False)

        print("Open dataset arguments:")

        if open_dataset_args:
            print()
            print(yaml.dump(open_dataset_args, indent=4, default_flow_style=False))

        if open_dataset_kwargs:
            print()
            print(yaml.dump(open_dataset_kwargs, indent=4, default_flow_style=False))

    def requirements(self, c: Checkpoint, args: Namespace) -> None:
        """Print the requirements for the checkpoint, including PyPI and Git dependencies.

        Parameters
        ----------
        c : Checkpoint
            The checkpoint object.
        args : Namespace
            The command-line arguments.
        """
        c = Checkpoint(args.path)
        r = c.provenance_training()

        f = sys.stdout

        if args.dump:
            print(json.dumps(r, indent=2, sort_keys=True), file=f)
            return

        print("# This file is automatically generated from a checkpoint.", file=f)
        print("# Checkpoint:", args.path, file=f)
        print("# Python:", r.get("python"), file=f)
        print(file=f)

        distribution_names = r.get("distribution_names", {})
        distribution_names.update(PACKAGES)

        pypi_requirements = {}
        git_requirements = {}

        for k, v in r.get("module_versions", {}).items():
            if k.startswith("_"):
                continue
            if v[0].isdigit():
                v = [x for x in v.split(".") if x.isdigit()]
                v = ".".join(v)
                k = k.replace(".", "-")
                pypi_requirements[distribution_names.get(k, k)] = v

        for k, v in r.get("git_versions", {}).items():
            if not k.startswith("anemoi."):
                continue

            sha1 = v.get("git", {}).get("sha1")

            if not sha1:
                continue

            what = k.split(".")[-1]
            if what in CORE:
                url = f"git+https://github.com/ecmwf/anemoi-core@{sha1}#subdirectory={what}"
            else:
                url = f"git+https://github.com/ecmwf/anemoi-{what}@{sha1}"

            k = k.replace(".", "-")
            git_requirements[k] = url

        if git_requirements:
            print(file=f)
            print("# Git requirements:", file=f)
            print(file=f)

        for k, v in sorted(git_requirements.items()):
            if k in SKIP:
                continue

            version = pypi_requirements.pop(k, None)
            if version:
                print(f"# {k}=={version}", file=f)
            print(v, file=f)

        if pypi_requirements:
            print(file=f)
            print("# PyPI requirements:", file=f)
            print(file=f)

        for k, v in sorted(pypi_requirements.items()):
            if k in SKIP:
                continue

            print(f"{k}=={v}", file=f)


command = InspectCmd
